# Copyright (c) 2019 - 2020 Simon Kern
# Copyright (c) 2015 Holger Nahrstaedt

import os, sys
import shutil
import gc
import numpy as np
# from numpy.testing import (assert_raises, run_module_suite,
#                            assert_equal, assert_allclose, assert_almost_equal)
import unittest
from pyedflib import highlevel
from pyedflib.edfwriter import EdfWriter
from pyedflib.edfreader import EdfReader
from datetime import datetime, date

class TestHighLevel(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        data_dir = os.path.join(os.path.dirname(__file__), 'data')

        cls.edfplus_data_file = os.path.join(data_dir, 'tmp_test_file_plus.edf')
        cls.test_generator = os.path.join(data_dir, 'test_generator.edf')
        cls.test_accented = os.path.join(data_dir, "tmp_áä'üöß.edf")
        cls.test_unicode = os.path.join(data_dir, "tmp_utf8-中文źąşㆆ운ʷᨄⅡəПр🤖.edf")
        cls.anonymized = os.path.join(data_dir, "tmp_anonymized.edf")
        cls.personalized = os.path.join(data_dir, "tmp_personalized.edf")
        cls.drop_from = os.path.join(data_dir, 'tmp_drop_from.edf')
        cls.tmp_testfile = os.path.join(data_dir, 'tmp')

    @classmethod
    def tearDownClass(cls):
        data_dir = os.path.join(os.path.dirname(__file__), 'data')
        tmpfiles = [f for f in os.listdir(data_dir) if f.startswith('tmp')]
        for file in tmpfiles:
            try:
                os.remove(os.path.join(data_dir, file))
            except Exception as e:
                print(e)

    def tearDown(self):
        # small hack to close handles in case of tests throwing an exception
        for obj in gc.get_objects():
            if isinstance(obj, (EdfWriter, EdfReader)):
                obj.close()
                del obj

    def test_dig2phys_calc(self):
        signals_phys, shead, _ = highlevel.read_edf(self.test_generator)
        signals_dig, _, _ = highlevel.read_edf(self.test_generator, digital=True)

        dmin, dmax = shead[0]['digital_min'],  shead[0]['digital_max']
        pmin, pmax = shead[0]['physical_min'],  shead[0]['physical_max']

        # convert to physical
        signal_phys2 = highlevel.dig2phys(signals_dig, dmin, dmax, pmin, pmax)
        np.testing.assert_allclose(signals_phys, signal_phys2)

        # convert to digital
        signals_dig2 = highlevel.phys2dig(signals_phys, dmin, dmax, pmin, pmax)
        signals_dig2 = np.rint(signals_dig2)
        np.testing.assert_allclose(signals_dig, signals_dig2)

    def test_read_write_edf(self):
        startdate = datetime.now()
        t = startdate
        startdate = datetime(t.year,t.month,t.day,t.hour, t.minute,t.second)

        header = highlevel.make_header(technician='tech', recording_additional='r_add',
                                                patientname='name', patient_additional='p_add',
                                                patientcode='42', equipment='eeg', admincode='120',
                                                sex='Male', startdate=startdate,birthdate='05.09.1980')
        annotations = [[0.01, -1, 'begin'],[0.5, -1, 'middle'],[10, -1, 'end']]

        signal_headers1 = highlevel.make_signal_headers(['ch'+str(i) for i in range(5)])

        for file_type in [-1,0,1,2,3]:
            if file_type in [0, 2]:
                header['annotations'] = []
            else:
                header['annotations'] = annotations

            file = f'{self.tmp_testfile}_{file_type}_phys.edf'
            signals = np.random.rand(5, 256*300)*200 #5 minutes of eeg
            success = highlevel.write_edf(file, signals, signal_headers1, header, file_type=file_type)
            self.assertTrue(os.path.isfile(file))
            self.assertGreater(os.path.getsize(file), 0)
            self.assertTrue(success)

            signals2, signal_headers2, header2 = highlevel.read_edf(file)

            self.assertEqual(len(signals2), 5)
            self.assertEqual(len(signals2), len(signal_headers2))
            for shead1, shead2 in zip(signal_headers1, signal_headers2):
                # When only 'sample_rate' is present, we use its value to write
                # the file, ignoring 'sample_frequency', which means that when
                # we read it back only the 'sample_rate' value is present.
                self.assertDictEqual({**shead1, 'sample_frequency': shead1['sample_rate']},
                                     shead2)
            np.testing.assert_allclose(signals, signals2, atol=0.01)
            if file_type in [-1, 1, 3]:
                self.assertDictEqual(header, header2)

            file = f'{self.tmp_testfile}_{file_type}_dig.edf'
            signals = (signals*100).astype(np.int8)
            success = highlevel.write_edf(file, signals,  signal_headers1, header, digital=True)
            self.assertTrue(os.path.isfile(file))
            self.assertGreater(os.path.getsize(file), 0)
            self.assertTrue(success)

            signals2, signal_headers2, header2 = highlevel.read_edf(file, digital=True)

            self.assertEqual(len(signals2), 5)
            self.assertEqual(len(signals2), len(signal_headers2))
            np.testing.assert_array_equal(signals, signals2)
            for shead1, shead2 in zip(signal_headers1, signal_headers2):
                # When only 'sample_rate' is present, we use its value to write
                # the file, ignoring 'sample_frequency', which means that when
                # we read it back only the 'sample_rate' value is present.
                self.assertDictEqual({**shead1, 'sample_frequency': shead1['sample_rate']},
                                     shead2)
            # EDF/BDF header writing does not correctly work yet
            if file_type in [-1, 1, 3]:
                self.assertDictEqual(header, header2)

    def test_read_write_with_annotations(self):
        signals, signal_headers, header = highlevel.read_edf(self.test_generator)
        expected = [[0.0, -1, 'Recording starts'], [600.0, -1, 'Recording ends']]
        self.assertEqual(header['annotations'], expected)

        highlevel.write_edf(self.edfplus_data_file, signals, signal_headers, header)
        signals2, signal_header2s, header2 = highlevel.read_edf(self.edfplus_data_file)
        self.assertEqual(header['annotations'], header2['annotations'])


    def test_quick_write(self):
        signals = np.random.randint(-2048, 2048, [3, 256*60])
        highlevel.write_edf_quick(self.edfplus_data_file, signals.astype(np.int32), sfreq=256, digital=True)
        signals2, _, _ = highlevel.read_edf(self.edfplus_data_file, digital=True, verbose=True)
        np.testing.assert_allclose(signals, signals2)
        signals = np.random.rand(3, 256*60) # then rescale to 0-1
        signals = (signals - signals.min()) / (signals.max() - signals.min())
        highlevel.write_edf_quick(self.edfplus_data_file, signals, sfreq=256)
        signals2, _, _ = highlevel.read_edf(self.edfplus_data_file)
        np.testing.assert_allclose(signals, signals2, atol=0.00002)

    def test_fortran_write(self):
        # Create Fortran contiguous array
        signals = np.random.randint(-2048,2048,[4, 5000000])
        signals = np.asfortranarray(signals)
        # Write
        highlevel.write_edf_quick(self.edfplus_data_file, signals.astype(np.int32), sfreq=250, digital=True)
        # Read and check
        signals2, _, _ = highlevel.read_edf(self.edfplus_data_file, digital=True, verbose=True)
        np.testing.assert_allclose(signals, signals2)

        # Create Fortran contiguous list
        signals = [np.random.randint(-2048,2048,(5000000,), dtype=np.int32)]*4
        # Write
        highlevel.write_edf_quick(self.edfplus_data_file, signals, sfreq=250, digital=True)
        # Read and check
        signals2, _, _ = highlevel.read_edf(self.edfplus_data_file, digital=True, verbose=True)
        np.testing.assert_allclose(signals, signals2)


    def test_read_write_decimal_sample_frequencies(self):
        # first test with digital signals
        signals = np.random.randint(-2048, 2048, [3, 256*60+8])
        highlevel.write_edf_quick(self.edfplus_data_file, signals.astype(np.int32), sfreq=8.5, digital=True)
        signals2, _, _ = highlevel.read_edf(self.edfplus_data_file, digital=True, verbose=True)
        np.testing.assert_allclose(signals, signals2)

        # now with physical signals
        signals = np.random.rand(3, 256*60+8) # then rescale to 0-1
        signals = (signals - signals.min()) / (signals.max() - signals.min())
        highlevel.write_edf_quick(self.edfplus_data_file, signals, sfreq=8.5, digital=False)
        signals2, _, _ = highlevel.read_edf(self.edfplus_data_file, digital=False, verbose=True)
        np.testing.assert_allclose(signals, signals2, atol=0.0001)


    def test_read_write_diff_sfreq(self):

        signals = []
        sfreqs = [1, 64, 128, 200]
        sheaders = []
        for sfreq in sfreqs:
            signals.append(np.random.randint(-2048, 2048, sfreq*60).astype(np.int32))
            shead = highlevel.make_signal_header(f'ch{sfreq}', sample_frequency=sfreq)
            sheaders.append(shead)
        highlevel.write_edf(self.edfplus_data_file, signals, sheaders, digital=True)
        signals2, sheaders2, _ = highlevel.read_edf(self.edfplus_data_file, digital=True)
        for s1, s2 in zip(signals, signals2):
            np.testing.assert_allclose(s1, s2)

    def test_assertion_dmindmax(self):

        # test digital and dmin wrong
        signals =[np.random.randint(-2048, 2048, 256*60).astype(np.int32)]
        sheaders = [highlevel.make_signal_header('ch1', sample_frequency=256)]
        sheaders[0]['digital_min'] = -128
        sheaders[0]['digital_max'] = 128
        with self.assertRaises(AssertionError):
            highlevel.write_edf(self.edfplus_data_file, signals, sheaders, digital=True)

        # test pmin wrong
        signals = [np.random.randint(-2048, 2048, 256*60)]
        sheaders = [highlevel.make_signal_header('ch1', sample_frequency=256)]
        sheaders[0]['physical_min'] = -200
        sheaders[0]['physical_max'] = 200
        with self.assertRaises(AssertionError):
            highlevel.write_edf(self.edfplus_data_file, signals, sheaders, digital=False)


    def test_read_write_accented(self):
        signals = np.random.rand(3, 256*60) # then rescale to 0-1
        signals = (signals - signals.min()) / (signals.max() - signals.min())
        highlevel.write_edf_quick(self.test_accented, signals, sfreq=256)
        signals2, _, _ = highlevel.read_edf(self.test_accented)

        np.testing.assert_allclose(signals, signals2, atol=0.00002)
        # if os.name!='nt':
        self.assertTrue(os.path.isfile(self.test_accented), 'File does not exist')

    def test_read_unicode(self):
        signals = np.random.rand(3, 256*60) # then rescale to 0-1
        signals = (signals - signals.min()) / (signals.max() - signals.min())
        success = highlevel.write_edf_quick(self.edfplus_data_file, signals, sfreq=256)
        self.assertTrue(success)
        shutil.copy(self.edfplus_data_file, self.test_unicode)
        signals2, _, _ = highlevel.read_edf(self.test_unicode)
        self.assertTrue(os.path.isfile(self.test_unicode), 'File does not exist')


    def test_read_header(self):

        header = highlevel.read_edf_header(self.test_generator)
        self.assertEqual(len(header), 15)
        self.assertEqual(len(header['channels']), 11)
        self.assertEqual(len(header['SignalHeaders']), 11)
        self.assertEqual(header['Duration'], 600)
        self.assertEqual(header['admincode'], 'Dr. X')
        self.assertEqual(header['birthdate'], '30 jun 1969')
        self.assertEqual(header['equipment'], 'test generator')
        self.assertEqual(header['sex'], 'Male')
        self.assertEqual(header['gender'], 'Male')
        self.assertEqual(header['patient_additional'], 'patient')
        self.assertEqual(header['patientcode'], 'abcxyz99')
        self.assertEqual(header['patientname'], 'Hans Muller')
        self.assertEqual(header['technician'], 'Mr. Spotty')


    def test_anonymize(self):

        header = highlevel.make_header(technician='tech', recording_additional='radd',
                                                patientname='name', patient_additional='padd',
                                                patientcode='42', equipment='eeg', admincode='420',
                                                sex='Male', birthdate='05.09.1980')
        annotations = [[0.01, -1, 'begin'],[0.5, -1, 'middle'],[10, -1, 'end']]
        header['annotations'] = annotations
        signal_headers = highlevel.make_signal_headers(['ch'+str(i) for i in range(3)])
        signals = np.random.rand(3, 256*300)*200 #5 minutes of eeg
        signals = (signals - signals.min()) / (signals.max() - signals.min())
        highlevel.write_edf(self.personalized, signals, signal_headers, header)



        highlevel.anonymize_edf(self.personalized, new_file=self.anonymized,
                                        to_remove=['patientname', 'birthdate',
                                                   'admincode', 'patientcode',
                                                   'technician'],
                                        new_values=['x', '', 'xx', 'xxx',
                                                    'xxxx'], verify=True, verbose=True)
        new_header = highlevel.read_edf_header(self.anonymized)
        self.assertEqual(new_header['birthdate'], '')
        self.assertEqual(new_header['patientname'], 'x')
        self.assertEqual(new_header['admincode'], 'xx')
        self.assertEqual(new_header['patientcode'], 'xxx')
        self.assertEqual(new_header['technician'], 'xxxx')


        highlevel.anonymize_edf(self.personalized, to_remove=['patientname', 'birthdate',
                                                   'admincode', 'patientcode',
                                                   'technician'],
                                        new_values=['x', '', 'xx', 'xxx',
                                                    'xxxx'], verify=True)
        new_header = highlevel.read_edf_header(self.personalized[:-4]+'_anonymized.edf')
        self.assertEqual(new_header['birthdate'], '')
        self.assertEqual(new_header['patientname'], 'x')
        self.assertEqual(new_header['admincode'], 'xx')
        self.assertEqual(new_header['patientcode'], 'xxx')
        self.assertEqual(new_header['technician'], 'xxxx')

        with self.assertRaises(AssertionError):
            highlevel.anonymize_edf(self.personalized,
                                    new_file=self.anonymized,
                                    to_remove=['patientname', 'birthdate',
                                               'admincode', 'patientcode',
                                               'technician'],
                                    new_values=['x', '', 'xx', 'xxx'],
                                    verify=True)

    def test_drop_channel(self):
        signal_headers = highlevel.make_signal_headers(['ch'+str(i) for i in range(5)])
        signals = np.random.rand(5, 256*300)*200 #5 minutes of eeg
        signals = (signals - signals.min()) / (signals.max() - signals.min())
        highlevel.write_edf(self.drop_from, signals, signal_headers)

        dropped = highlevel.drop_channels(self.drop_from, to_keep=['ch1', 'ch2'], verbose=True)

        signals2, signal_headers, header = highlevel.read_edf(dropped)

        np.testing.assert_allclose(signals[1:3,:], signals2, atol=0.01)

        highlevel.drop_channels(self.drop_from, self.drop_from[:-4]+'2.edf',
                                to_drop=['ch0', 'ch1', 'ch2'])
        signals2, signal_headers, header = highlevel.read_edf(self.drop_from[:-4]+'2.edf')

        np.testing.assert_allclose(signals[3:,:], signals2, atol=0.01)

        with self.assertRaises(AssertionError):
            highlevel.drop_channels(self.drop_from, to_keep=['ch1'], to_drop=['ch3'])


    def test_annotation_bytestring(self):
        header = highlevel.make_header(technician='tech', recording_additional='radd',
                                                patientname='name', patient_additional='padd',
                                                patientcode='42', equipment='eeg', admincode='420',
                                                sex='Male', birthdate='05.09.1980')
        annotations = [[0.01, b'-1', 'begin'],[0.5, b'-1', 'middle'],[10, -1, 'end']]
        header['annotations'] = annotations
        signal_headers = highlevel.make_signal_headers(['ch'+str(i) for i in range(3)])
        signals = np.random.rand(3, 256*300)*200 #5 minutes of eeg
        highlevel.write_edf(self.edfplus_data_file, signals, signal_headers, header)
        _,_,header2 = highlevel.read_edf(self.edfplus_data_file)
        highlevel.write_edf(self.edfplus_data_file, signals, signal_headers, header)
        _,_,header3 = highlevel.read_edf(self.edfplus_data_file)
        self.assertEqual(header2['annotations'], header3['annotations'])


if __name__ == '__main__':
    # run_module_suite(argv=sys.argv)
    unittest.main()



# Copyright (c) 2016 Holger Nahrstaedt

import os
import numpy as np
from scipy import signal
import pyedflib

# signal label/waveform  amplitude    f       sf
# ---------------------------------------------------
#    1    squarewave        100 uV    0.1Hz   200 Hz
#    2    ramp              100 uV    1 Hz    200 Hz
#    3    pulse 1           100 uV    1 Hz    200 Hz
#    4    pulse 2           100 uV    1 Hz    256 Hz
#    5    pulse 3           100 uV    1 Hz    217 Hz
#    6    noise             100 uV    - Hz    200 Hz
#    7    sine 1 Hz         100 uV    1 Hz    200 Hz
#    8    sine 8 Hz         100 uV    8 Hz    200 Hz
#    9    sine 8.1777 Hz    100 uV    8.25 Hz 200 Hz
#    10    sine 8.5 Hz       100 uV    8.5Hz   200 Hz
#    11    sine 15 Hz        100 uV   15 Hz    200 Hz
#    12    sine 17 Hz        100 uV   17 Hz    200 Hz
#    13    sine 50 Hz        100 uV   50 Hz    200 Hz


if __name__ == '__main__':
    test_data_file = os.path.join('.', 'test_generator2.edf')
    file_duration = 600
    f = pyedflib.EdfWriter(test_data_file, 13,
                           file_type=pyedflib.FILETYPE_EDFPLUS)
    channel_info = []
    data_list = []

    ch_dict = {'label': 'squarewave', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*200)
    xtemp = np.sin(2*np.pi*0.1*time)
    x1 = xtemp.copy()
    x1[np.where(xtemp > 0)[0]] = 100
    x1[np.where(xtemp < 0)[0]] = -100
    data_list.append(x1)

    ch_dict = {'label': 'ramp', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*200)
    x2 = signal.sawtooth(2 * np.pi * 1 * time)
    data_list.append(x2)

    ch_dict = {'label': 'pulse 1', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*200)
    xtemp = np.sin(2*np.pi*0.5*time)
    x3 = np.zeros(file_duration*200)
    x3[np.where(np.all((xtemp < 0.02, xtemp > -0.02),axis=0))[0]] = 100
    data_list.append(x3)

    ch_dict = {'label': 'pulse 2', 'dimension': 'uV', 'sample_frequency': 256, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*256)
    xtemp = np.sin(2*np.pi*0.5*time)
    x4 = np.zeros(file_duration*256)
    x4[np.where(np.all((xtemp < 0.02, xtemp > -0.02),axis=0))[0]] = 100
    data_list.append(x4)

    ch_dict = {'label': 'pulse 3', 'dimension': 'uV', 'sample_frequency': 217, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*217)
    xtemp = np.sin(2*np.pi*0.5*time)
    x5 = np.zeros(file_duration*217)
    x5[np.where(np.all((xtemp < 0.02, xtemp > -0.02),axis=0))[0]] = 100
    data_list.append(x5)

    ch_dict = {'label': 'noise', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    data_list.append(np.random.normal(size=file_duration*200))

    ch_dict = {'label': 'sine 1 Hz', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*200)
    data_list.append(np.sin(2*np.pi*1*time))

    ch_dict = {'label': 'sine 8 Hz', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*200)
    data_list.append(np.sin(2*np.pi*8*time))

    ch_dict = {'label': 'sine 8.1777 Hz', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*200)
    data_list.append(np.sin(2*np.pi*8.1777*time))

    ch_dict = {'label': 'sine 8.5 Hz', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*200)
    data_list.append(np.sin(2*np.pi*8.5*time))

    ch_dict = {'label': 'sine 15 Hz', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*200)
    data_list.append(np.sin(2*np.pi*15*time))

    ch_dict = {'label': 'sine 17 Hz', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*200)
    data_list.append(np.sin(2*np.pi*17*time))

    ch_dict = {'label': 'sine 50 Hz', 'dimension': 'uV', 'sample_frequency': 200, 'physical_max': 100, 'physical_min': -100, 'digital_max': 32767, 'digital_min': -32768, 'transducer': '', 'prefilter':''}
    channel_info.append(ch_dict)
    time = np.linspace(0, file_duration, file_duration*200)
    data_list.append(np.sin(2*np.pi*50*time))

    f.setSignalHeaders(channel_info)
    f.writeSamples(data_list)
    f.writeAnnotation(0, -1, "Recording starts")
    f.writeAnnotation(298, -1, "Test 1")
    f.writeAnnotation(294.99, -1, "pulse 1")
    f.writeAnnotation(295.9921875, -1, "pulse 2")
    f.writeAnnotation(296.99078341013825, -1, "pulse 3")
    f.writeAnnotation(600, -1, "Recording ends")
    f.close()
    del f



# Copyright (c) 2015 - 2017 Holger Nahrstaedt
# Copyright (c) 2011, 2015, Chris Lee-Messer
# See LICENSE for license details.

# get constants
include "edf.pxi"

cdef extern from "c/edflib.h":
    int edf_set_patientcode(int, char *)
    int edfwrite_annotation_latin1(int, long long int, long long int, char *)
    int edfwrite_annotation_utf8(int, long long int, long long int, char *)
    int edflib_version()
    cdef struct edf_annotation_struct:
        long long int onset
        char * duration
        char * annotation
    int edf_get_annotation(int, int, edf_annotation_struct *)
    int edf_set_technician(int, char *)
    int edfread_digital_samples(int, int, int, int *)
    int edf_blockwrite_physical_samples(int, double *)
    int edf_set_recording_additional(int, char *)
    int edfwrite_physical_samples(int, double *)
    int edfclose_file(int)
    int edf_set_patientname(int, char *)
    int edf_set_physical_minimum(int, int, double)
    int edfread_physical_samples(int, int, int, double *)
    int edf_set_physical_maximum(int, int, double)
    int edfopen_file_writeonly(char *, int, int)
    int edf_set_patient_additional(int, char *)
    int edf_set_digital_maximum(int, int, int)
    cdef struct edf_param_struct:
        char  label[17]
        long long int smp_in_file
        double phys_max
        double phys_min
        int dig_max
        int dig_min
        int smp_in_datarecord
        char physdimension[9]
        char prefilter[81]
        char transducer[81]
        
    cdef struct edf_hdr_struct:
        int handle
        int filetype
        int edfsignals
        long long int file_duration
        int startdate_day
        int startdate_month
        int startdate_year
        int starttime_second
        int starttime_minute
        int starttime_hour
        long long starttime_subsecond
        char * patient
        char * recording
        char * patientcode
        char * gender
        char * birthdate
        char * patient_name
        char * patient_additional
        char * admincode
        char * technician
        char * equipment
        char * recording_additional
        long long int datarecord_duration
        long long int datarecords_in_file
        long long int annotations_in_file
        edf_param_struct * signalparam
    
    int edfopen_file_readonly(char *, edf_hdr_struct *, int, int)
    int edf_set_birthdate(int, int, int, int)
    int edf_set_digital_minimum(int, int, int)
    int edfwrite_digital_samples(int, int *)
    int edf_set_equipment(int, char *)
    int edf_set_samplefrequency(int, int, int)
    int edf_set_admincode(int, char *)
    int edf_set_label(int, int, char *)
    int edf_blockwrite_digital_samples(int, int *)
    long long int edftell(int, int)
    void edfrewind(int, int)
    int edf_set_gender(int, int)
    int edf_set_physical_dimension(int, int, char *)
    int edf_set_transducer(int, int, char *)
    int edf_set_prefilter(int, int, char *)
    long long int edfseek(int, int, long long int, int)
    int edf_set_startdatetime(int, int, int, int, int, int, int)
    int edf_set_subsecond_starttime(int, int)
    int edf_set_datarecord_duration(int, int)
    int edf_set_number_of_annotation_signals(int, int)

    # new functions in 1.10
    int edflib_is_file_used(const char *)
    int edflib_get_number_of_open_files()
    int edflib_get_handle(int)
    int edfwrite_digital_short_samples(int , short *)
    int edf_blockwrite_digital_short_samples(int, short *)
    int edf_blockwrite_digital_3byte_samples(int , void *)



# Copyright (c) 2015, Holger Nahrstaedt
# Copyright (c) 2011, 2015, Chris Lee-Messer
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#/* values for annotations */
DEF EDFLIB_DO_NOT_READ_ANNOTATIONS = 0
DEF EDFLIB_READ_ANNOTATIONS       = 1
DEF EDFLIB_READ_ALL_ANNOTATIONS   = 2

DEF EDFLIB_CHECK_FILE_SIZE = 0
DEF EDFLIB_DO_NOT_CHECK_FILE_SIZE       = 1
DEF EDFLIB_REPAIR_FILE_SIZE_IF_WRONG   = 2

DEF EDFLIB_TIME_DIMENSION = 10000000LL
DEF EDFLIB_MAXSIGNALS = 256
DEF EDFLIB_MAX_ANNOTATION_LEN = 512

DEF EDFSEEK_SET = 0
DEF EDFSEEK_CUR = 1
DEF EDFSEEK_END = 2
#/* the following defines are used in the member "filetype" of the edf_hdr_struct */
#/* and as return value for the function edfopen_file_readonly() */
DEF EDFLIB_FILETYPE_EDF =                  0
DEF EDFLIB_FILETYPE_EDFPLUS =              1
DEF EDFLIB_FILETYPE_BDF    =              2
DEF EDFLIB_FILETYPE_BDFPLUS  =             3
DEF EDFLIB_MALLOC_ERROR       =          -1
DEF EDFLIB_NO_SUCH_FILE_OR_DIRECTORY =   -2
DEF EDFLIB_FILE_CONTAINS_FORMAT_ERRORS = -3
DEF EDFLIB_MAXFILES_REACHED         =    -4
DEF EDFLIB_FILE_READ_ERROR           =   -5
DEF EDFLIB_FILE_ALREADY_OPENED       =   -6
DEF EDFLIB_FILETYPE_ERROR            =   -7
DEF EDFLIB_FILE_WRITE_ERROR          =   -8
DEF EDFLIB_NUMBER_OF_SIGNALS_INVALID =   -9
DEF EDFLIB_FILE_IS_DISCONTINUOUS     =  -10
DEF EDFLIB_INVALID_READ_ANNOTS_VALUE =  -11
DEF EDFLIB_INVALID_CHECK_SIZE_VALUE  =  -12


#/* the following defines are possible errors returned by edfopen_file_writeonly() */
DEF EDFLIB_NO_SIGNALS              =    -20
DEF EDFLIB_TOO_MANY_SIGNALS         =   -21
DEF EDFLIB_NO_SAMPLES_IN_RECORD      =  -22
DEF EDFLIB_DIGMIN_IS_DIGMAX           = -23
DEF EDFLIB_DIGMAX_LOWER_THAN_DIGMIN   = -24
DEF EDFLIB_PHYSMIN_IS_PHYSMAX         = -25
DEF EDFLIB_DATARECORD_SIZE_TOO_BIG    = -26
DEF EDFLIB_FILE_ERRORS_STARTDATE      =-30
DEF EDFLIB_FILE_ERRORS_STARTTIME      =-31
DEF EDFLIB_FILE_ERRORS_NUMBER_SIGNALS =-32
DEF EDFLIB_FILE_ERRORS_BYTES_HEADER   =-33
DEF EDFLIB_FILE_ERRORS_RESERVED_FIELD =-34
DEF EDFLIB_FILE_ERRORS_NUMBER_DATARECORDS =-35
DEF EDFLIB_FILE_ERRORS_DURATION =-36
DEF EDFLIB_FILE_ERRORS_LABEL =-37
DEF EDFLIB_FILE_ERRORS_TRANSDUCER =-38
DEF EDFLIB_FILE_ERRORS_PHYS_DIMENSION =-39
DEF EDFLIB_FILE_ERRORS_PHYS_MAX =-40
DEF EDFLIB_FILE_ERRORS_PHYS_MIN =-41
DEF EDFLIB_FILE_ERRORS_DIG_MAX =-42
DEF EDFLIB_FILE_ERRORS_DIG_MIN =-43
DEF EDFLIB_FILE_ERRORS_PREFILTER =-44
DEF EDFLIB_FILE_ERRORS_SAMPLES_DATARECORD =-45
DEF EDFLIB_FILE_ERRORS_FILESIZE =-46
DEF EDFLIB_FILE_ERRORS_RECORDINGFIELD =-47
DEF EDFLIB_FILE_ERRORS_PATIENTNAME =-48


# Copyright (c) 2019 - 2023 Simon Kern
# Copyright (c) 2015 - 2023 Holger Nahrstaedt
# Copyright (c) 2011, 2015, Chris Lee-Messer
# Copyright (c) 2016-2017 The pyedflib Developers
#                         <https://github.com/holgern/pyedflib>
# See LICENSE for license details.
"""
Created on Tue Jan  7 12:13:47 2020

This file contains high-level functions to work with pyedflib.

Includes
    - Reading and writing EDFs
    - Anonymizing EDFs
    - Comparing EDFs
    - Renaming Channels from EDF files
    - Dropping Channels from EDF files

@author: skjerns
"""

import os
import numpy as np
import warnings
import pyedflib
from copy import deepcopy
from datetime import datetime
# from . import EdfWriter
# from . import EdfReader


def _get_sample_frequency(signal_header):
    # Temporary conditional assignment while we deprecate 'sample_rate' as a channel attribute
    # in favor of 'sample_frequency', supporting the use of either to give
    # users time to switch to the new interface.
    return (signal_header['sample_rate']
            if signal_header.get('sample_frequency') is None
            else signal_header['sample_frequency'])


def tqdm(iterable, *args, **kwargs):
    """
    These is an optional dependency that shows a progress bar for some
    of the functions, e.g. loading.

    install this dependency with `pip install tqdm`

    if not installed this is just a pass through iterator.
    """
    try:
        from tqdm import tqdm as iterator
        return iterator(iterable, *args, **kwargs)
    except:
        return iterable


def _parse_date(string):
    """
    A simple dateparser that detects common  date formats

    Parameters
    ----------
    string : str
        a date string in format as denoted below.

    Returns
    -------
    datetime.datetime
        datetime object of a time.

    """
    # some common formats.
    formats = ['%Y-%m-%d', '%d-%m-%Y', '%d.%m.%Y', '%Y.%m.%d', '%d %b %Y',
               '%Y/%m/%d', '%d/%m/%Y']
    for f in formats:
        try:
            return datetime.strptime(string, f)
        except:
            pass
    try:
        import dateparser
        return dateparser.parse(string)
    except:
        print('dateparser is not installed. to convert strings to dates'\
              'install via `pip install dateparser`.')
        raise ValueError('birthdate must be datetime object or of format'\
                         ' `%d-%m-%Y`, eg. `24-01-2020`')

def dig2phys(signal, dmin, dmax, pmin, pmax):
    """
    converts digital edf values to physical values

    Parameters
    ----------
    signal : np.ndarray or int
        A numpy array with int values (digital values) or an int.
    dmin : int
        digital minimum value of the edf file (eg -2048).
    dmax : int
        digital maximum value of the edf file (eg 2048).
    pmin : float
        physical maximum value of the edf file (eg -200.0).
    pmax : float
        physical maximum value of the edf file (eg 200.0).

    Returns
    -------
    physical : np.ndarray or float
        converted physical values

    """
    m = (pmax-pmin) / (dmax-dmin)
    b = pmax / m - dmax
    physical = m * (signal + b)
    return physical


def phys2dig(signal, dmin, dmax, pmin, pmax):
    """
    converts physical values to digital values

    Parameters
    ----------
    signal : np.ndarray or int
        A numpy array with int values (digital values) or an int.
    dmin : int
        digital minimum value of the edf file (eg -2048).
    dmax : int
        digital maximum value of the edf file (eg 2048).
    pmin : float
        physical maximum value of the edf file (eg -200.0).
    pmax : float
        physical maximum value of the edf file (eg 200.0).

    Returns
    -------
    digital : np.ndarray or int
        converted digital values

    """
    m = (pmax-pmin) / (dmax-dmin)
    b = pmax / m - dmax
    digital = signal/m - b
    return digital



def make_header(technician='', recording_additional='', patientname='',
                patient_additional='', patientcode= '', equipment= '',
                admincode= '', sex= '', startdate=None, birthdate= '',
                gender=None):
    """
    A convenience function to create an EDF header (a dictionary) that
    can be used by pyedflib to update the main header of the EDF

    Parameters
    ----------
    technician : str, optional
        name of the technician. The default is ''.
    recording_additional : str, optional
        comments etc. The default is ''.
    patientname : str, optional
        the name of the patient. The default is ''.
    patient_additional : TYPE, optional
        more info about the patient. The default is ''.
    patientcode : str, optional
        alphanumeric code. The default is ''.
    equipment : str, optional
        which system was used. The default is ''.
    admincode : str, optional
        code of the admin. The default is ''.
    sex : str, optional
        sex of patient. The default is ''.
    startdate : datetime.datetime, optional
        startdate of recording. The default is None.
    birthdate : str/datetime.datetime, optional
        date of birth of the patient. The default is ''.

    Returns
    -------
    header : dict
        a dictionary with the values given filled in.

    """

    if not birthdate=='' and isinstance(birthdate, str):
        birthdate = _parse_date(birthdate)
    if startdate is None:
        now = datetime.now()
        startdate = datetime(now.year, now.month, now.day,
                             now.hour, now.minute, now.second)
        del now
    if isinstance(birthdate, datetime):
        birthdate = birthdate.strftime('%d %b %Y').lower()

    # backwards compatibility
    if gender is not None:
        if sex == '':
            sex = gender
            warnings.warn("Parameter 'gender' is deprecated, use 'sex' instead.", DeprecationWarning, stacklevel=2)
        elif sex != gender:
            raise ValueError("Defined both parameters 'sex' and 'gender', with different values: {sex} != {gender}")
    gender = sex

    local = locals()

    header = {}
    for var in local:
        if isinstance(local[var], datetime):
            header[var] = local[var]
        else:
            header[var] = str(local[var])
    return header


def make_signal_header(label, dimension='uV', sample_rate=256, sample_frequency=None,
                       physical_min=-200, physical_max=200, digital_min=-32768,
                       digital_max=32767, transducer='', prefiler=''):
    """
    A convenience function that creates a signal header for a given signal.
    This can be used to create a list of signal headers that is used by
    pyedflib to create an edf. With this, different sampling frequencies
    can be indicated.

    Parameters
    ----------
    label : str
        the name of the channel.
    dimension : str, optional
        dimension, eg mV. The default is 'uV'.
    sample_rate : int, optional
        sampling frequency. The default is 256. Deprecated: use 'sample_frequency' instead.
    sample_frequency : int, optional
        sampling frequency. The default is 256.
    physical_min : float, optional
        minimum value in dimension. The default is -200.
    physical_max : float, optional
        maximum value in dimension. The default is 200.
    digital_min : int, optional
        digital minimum of the ADC. The default is -32768.
    digital_max : int, optional
        digital maximum of the ADC. The default is 32767.
    transducer : str, optional
        electrode type that was used. The default is ''.
    prefiler : str, optional
        filtering and sampling method. The default is ''.

    Returns
    -------
    signal_header : dict
        a signal header that can be used to save a channel to an EDF.

    """

    signal_header = {'label': label,
               'dimension': dimension,
               'sample_rate': sample_rate,
               'sample_frequency': sample_frequency,
               'physical_min': physical_min,
               'physical_max': physical_max,
               'digital_min':  digital_min,
               'digital_max':  digital_max,
               'transducer': transducer,
               'prefilter': prefiler}
    return signal_header


def make_signal_headers(list_of_labels, dimension='uV', sample_rate=256,
                       sample_frequency=None, physical_min=-200.0, physical_max=200.0,
                       digital_min=-32768, digital_max=32767,
                       transducer='', prefiler=''):
    """
    A function that creates signal headers for a given list of channel labels.
    This can only be used if each channel has the same sampling frequency

    Parameters
    ----------
    list_of_labels : list of str
        A list with labels for each channel.
    dimension : str, optional
        dimension, eg mV. The default is 'uV'.
    sample_rate : int, optional
        sampling frequency. The default is 256.  Deprecated: use 'sample_frequency' instead.
    sample_frequency : int, optional
        sampling frequency. The default is 256.
    physical_min : float, optional
        minimum value in dimension. The default is -200.
    physical_max : float, optional
        maximum value in dimension. The default is 200.
    digital_min : int, optional
        digital minimum of the ADC. The default is -32768.
    digital_max : int, optional
        digital maximum of the ADC. The default is 32767.
    transducer : str, optional
        electrode type that was used. The default is ''.
    prefiler : str, optional
        filtering and sampling method. The default is ''.

    Returns
    -------
    signal_headers : list of dict
        returns n signal headers as a list to save several signal headers.

    """
    signal_headers = []
    for label in list_of_labels:
        header = make_signal_header(label, dimension=dimension, sample_rate=sample_rate,
                                    sample_frequency=sample_frequency,
                                    physical_min=physical_min, physical_max=physical_max,
                                    digital_min=digital_min, digital_max=digital_max,
                                    transducer=transducer, prefiler=prefiler)
        signal_headers.append(header)
    return signal_headers


def read_edf(edf_file, ch_nrs=None, ch_names=None, digital=False, verbose=False):
    """
    Convenience function for reading EDF+/BDF data with pyedflib.

    Will load the edf and return the signals, the headers of the signals
    and the header of the EDF. If all signals have the same sample frequency
    will return a numpy array, else a list with the individual signals


    Parameters
    ----------
    edf_file : str
        link to an edf file.
    ch_nrs : list of int, optional
        The indices of the channels to read. The default is None.
    ch_names : list of str, optional
        The names of channels to read. The default is None.
    digital : bool, optional
        will return the signals as digital values (ADC). The default is False.
    verbose : bool, optional
        Print progress bar while loading or not. The default is False.

    Returns
    -------
    signals : np.ndarray or list
        the signals of the chosen channels contained in the EDF.
    signal_headers : list
        one signal header for each channel in the EDF.
    header : dict
        the main header of the EDF file containing meta information.

    """
    assert (ch_nrs is  None) or (ch_names is None), \
           'names xor numbers should be supplied'
    if ch_nrs is not None and not isinstance(ch_nrs, list): ch_nrs = [ch_nrs]
    if ch_names is not None and \
        not isinstance(ch_names, list): ch_names = [ch_names]

    with pyedflib.EdfReader(edf_file) as f:
        # see which channels we want to load
        available_chs = [ch.upper() for ch in f.getSignalLabels()]
        n_chrs = f.signals_in_file

        # find out which number corresponds to which channel
        if ch_names is not None:
            ch_nrs = []
            for ch in ch_names:
                if not ch.upper() in available_chs:
                    warnings.warn('{} is not in source file (contains {})'\
                                  .format(ch, available_chs))
                    print('will be ignored.')
                else:
                    ch_nrs.append(available_chs.index(ch.upper()))

        # if there ch_nrs is not given, load all channels

        if ch_nrs is None: # no numbers means we load all
            ch_nrs = range(n_chrs)

        # convert negative numbers into positives
        ch_nrs = [n_chrs+ch if ch<0 else ch for ch in ch_nrs]

        # load headers, signal information and
        header = f.getHeader()
        signal_headers = [f.getSignalHeaders()[c] for c in ch_nrs]

        # add annotations to header
        annotations = f.readAnnotations()
        annotations = [[s, d, a] for s,d,a in zip(*annotations)]
        header['annotations'] = annotations


        signals = []
        for i,c in enumerate(tqdm(ch_nrs, desc='Reading Channels',
                                  disable=not verbose)):
            signal = f.readSignal(c, digital=digital)
            signals.append(signal)

        # we can only return a np.array if all signals have the same samplefreq
        sfreqs = [_get_sample_frequency(shead) for shead in signal_headers]
        all_sfreq_same = sfreqs[1:]==sfreqs[:-1]
        if all_sfreq_same:
            dtype = np.int32 if digital else float
            signals = np.array(signals, dtype=dtype)

    assert len(signals)==len(signal_headers), 'Something went wrong, lengths'\
                                         ' of headers is not length of signals'
    del f
    return  signals, signal_headers, header


def write_edf(edf_file, signals, signal_headers, header=None, digital=False,
              file_type=-1):
    """
    Write signals to an edf_file. Header can be generated on the fly with
    generic values. EDF+/BDF+ is selected based on the filename extension,
    but can be overwritten by setting file_type to pyedflib.FILETYPE_XXX

    Parameters
    ----------
    edf_file : np.ndarray or list
        where to save the EDF file
    signals : list
        The signals as a list of arrays or a ndarray.

    signal_headers : list of dict
        a list with one signal header(dict) for each signal.
        See pyedflib.EdfWriter.setSignalHeader..
    header : dict
        a main header (dict) for the EDF file, see
        pyedflib.EdfWriter.setHeader for details.
        If no header present, will create an empty header
    digital : bool, optional
        whether the signals are in digital format (ADC). The default is False.
    file_type: int, optional
        choose file_type for saving.
        EDF = 0, EDF+ = 1, BDF = 2, BDF+ = 3, automatic from extension = -1

    Returns
    -------
    bool
         True if successful, False if failed.
    """
    assert header is None or isinstance(header, dict), \
        'header must be dictioniary or None'
    assert isinstance(signal_headers, list), \
        'signal headers must be list'
    assert len(signal_headers)==len(signals), \
        'signals and signal_headers must be same length'
    assert file_type in [-1, 0, 1, 2, 3], \
        'file_type must be in range -1, 3'

    # copy objects to prevent accidental changes to mutable objects
    header = deepcopy(header)
    signal_headers = deepcopy(signal_headers)

    if file_type==-1:
        ext = os.path.splitext(edf_file)[-1]
        if ext.lower() == '.edf':
            file_type = pyedflib.FILETYPE_EDFPLUS
        elif ext.lower() == '.bdf':
            file_type = pyedflib.FILETYPE_BDFPLUS
        else:
            raise ValueError(f'Unknown extension {ext}')

    n_channels = len(signals)

    # if there is no header, we create one with dummy values
    if header is None:
        header = {}
    default_header = make_header()
    default_header.update(header)
    header = default_header

    # check dmin, dmax and pmin, pmax dont exceed signal min/max
    for sig, shead in zip(signals, signal_headers):
        dmin, dmax = shead['digital_min'], shead['digital_max']
        pmin, pmax = shead['physical_min'], shead['physical_max']
        label = shead['label']
        if digital: # exception as it will lead to clipping
            assert dmin<=sig.min(), \
            'digital_min is {}, but signal_min is {}' \
            'for channel {}'.format(dmin, sig.min(), label)
            assert dmax>=sig.max(), \
            'digital_min is {}, but signal_min is {}' \
            'for channel {}'.format(dmax, sig.max(), label)
            assert pmin != pmax, \
            f'physical_min {pmin} should be different from physical_max {pmax}'
        else: # only warning, as this will not lead to clipping
            assert pmin<=sig.min(), \
            'phys_min is {}, but signal_min is {} ' \
            'for channel {}'.format(pmin, sig.min(), label)
            assert pmax>=sig.max(), \
            'phys_max is {}, but signal_max is {} ' \
            'for channel {}'.format(pmax, sig.max(), label)

    # get annotations, in format [[timepoint, duration, description], [...]]
    annotations = header.get('annotations', [])

    with pyedflib.EdfWriter(edf_file, n_channels=n_channels, file_type=file_type) as f:
        f.setSignalHeaders(signal_headers)
        f.setHeader(header)
        f.writeSamples(signals, digital=digital)
        for annotation in annotations:
            f.writeAnnotation(*annotation)
    del f

    return os.path.isfile(edf_file)


def write_edf_quick(edf_file, signals, sfreq, digital=False):
    """
    wrapper for write_pyedf without creating headers.
    Use this if you don't care about headers or channel names and just
    want to dump some signals with the same sampling freq. to an edf

    Parameters
    ----------
    edf_file : str
        where to store the data/edf.
    signals : np.ndarray
        The signals you want to store as numpy array.
    sfreq : int
        the sampling frequency of the signals.
    digital : bool, optional
        if the data is present digitally (int) or as mV/uV.The default is False.

    Returns
    -------
    bool
        True if successful, else False or raise Error.

    """
    signals = np.atleast_2d(signals)
    header = make_header(technician='pyedflib-quickwrite')
    labels = [f'CH_{i}' for i in range(len(signals))]
    pmin, pmax = signals.min(), signals.max()
    signal_headers = make_signal_headers(labels, sample_frequency=sfreq,
                                         physical_min=pmin, physical_max=pmax)
    return write_edf(edf_file, signals, signal_headers, header, digital=digital)


def read_edf_header(edf_file, read_annotations=True):
    """
    Reads the header and signal headers of an EDF file and it's annotations

    Parameters
    ----------
    edf_file : str
        EDF/BDF file to read.

    Returns
    -------
    summary : dict
        header of the edf file as dictionary.

    """
    assert os.path.isfile(edf_file), f'file {edf_file} does not exist'
    with pyedflib.EdfReader(edf_file) as f:

        summary = f.getHeader()
        summary['Duration'] = f.getFileDuration()
        summary['SignalHeaders'] = f.getSignalHeaders()
        summary['channels'] = f.getSignalLabels()
        if read_annotations:
            annotations = f.read_annotation()
            annotations = [[float(t)/10000000, d if d else -1, x.decode()] for t,d,x in annotations]
            summary['annotations'] = annotations
    del f
    return summary


def compare_edf(edf_file1, edf_file2, verbose=False):
    """
    Loads two edf files and checks whether the values contained in
    them are the same. Does not check the header or annotations data.

    Mainly to verify that other options (eg anonymization) produce the
    same EDF file.

    Parameters
    ----------
    edf_file1 : str
        edf file 1 to compare.
    edf_file2 : str
        edf file 2 to compare.
    verbose : bool, optional
        print progress or not. The default is False.

    Returns
    -------
    bool
        True if signals are equal, else raises error.
    """
    signals1, shead1, _ =  read_edf(edf_file1, digital=True, verbose=verbose)
    signals2, shead2, _ =  read_edf(edf_file2, digital=True, verbose=verbose)

    for i, sigs in enumerate(zip(signals1, signals2)):
        s1, s2 = sigs
        if np.array_equal(s1, s2): continue # early stopping
        s1 = np.abs(s1)
        s2 = np.abs(s2)
        if np.array_equal(s1, s2): continue # early stopping
        close =  np.mean(np.isclose(s1, s2))
        assert close>0.99, 'Error, digital values of {}'\
              ' and {} for ch {}: {} are not the same: {:.3f}'.format(
                edf_file1, edf_file2, shead1[i]['label'],
                shead2[i]['label'], close)

    dmin1, dmax1 = shead1[i]['digital_min'], shead1[i]['digital_max']
    pmin1, pmax1 = shead1[i]['physical_min'], shead1[i]['physical_max']
    dmin2, dmax2 = shead2[i]['digital_min'], shead2[i]['digital_max']
    pmin2, pmax2 = shead2[i]['physical_min'], shead2[i]['physical_max']

    for i, sigs in enumerate(zip(signals1, signals2)):
        s1, s2 = sigs

        # convert to physical values, no need to load all data again
        s1 = dig2phys(s1, dmin1, dmax1, pmin1, pmax1)
        s2 = dig2phys(s2, dmin2, dmax2, pmin2, pmax2)

        # compare absolutes in case of inverted signals
        if np.array_equal(s1, s2): continue # early stopping
        s1 = np.abs(s1)
        s2 = np.abs(s2)
        if np.array_equal(s1, s2): continue # early stopping
        min_dist = np.abs(dig2phys(1, dmin1, dmax1, pmin1, pmax1))
        close =  np.mean(np.isclose(s1, s2, atol=min_dist))
        assert close>0.99, 'Error, physical values of {}'\
            ' and {} for ch {}: {} are not the same: {:.3f}'.format(
                edf_file1, edf_file2, shead1[i]['label'],
                shead2[i]['label'], close)
    return True


def drop_channels(edf_source, edf_target=None, to_keep=None, to_drop=None,
                  verbose=False):
    """
    Remove channels from an edf file. Save the file.
    For safety reasons, no source files can be overwritten.

    Parameters
    ----------
    edf_source : str
        The source edf file from which to drop channels.
    edf_target : str, optional
        Where to save the file.If None, will be edf_source+'dropped.edf'.
        The default is None.
    to_keep : list, optional
         A list of channel names or indices that will be kept.
         Strings will always be interpreted as channel names.
         'to_keep' will overwrite any droppings proposed by to_drop.
         The default is None.
    to_drop : list, optional
        A list of channel names/indices that should be dropped.
        Strings will be interpreted as channel names. The default is None.
    verbose : bool, optional
        print progress or not. The default is False.

    Returns
    -------
    edf_target : str
         the target filename with the dropped channels.

    """

    # convert to list if necessary
    if isinstance(to_keep, (int, str)): to_keep = [to_keep]
    if isinstance(to_drop, (int, str)): to_drop = [to_drop]

    # check all parameters are good
    assert to_keep is None or to_drop is None,'Supply only to_keep xor to_drop'
    if to_keep is not None:
        assert all([isinstance(ch, (str, int)) for ch in to_keep]),\
            'channels must be int or string'
    if to_drop is not None:
        assert all([isinstance(ch, (str, int)) for ch in to_drop]),\
            'channels must be int or string'
    assert os.path.exists(edf_source), \
            f'source file {edf_source} does not exist'
    assert edf_source!=edf_target, 'For safet, target must not be source file.'

    if edf_target is None:
        edf_target = os.path.splitext(edf_source)[0] + '_dropped.edf'
    if os.path.exists(edf_target):
        warnings.warn('Target file will be overwritten')

    ch_names = read_edf_header(edf_source)['channels']
    # convert to all lowercase for compatibility
    ch_names = [ch.lower() for ch in ch_names]
    ch_nrs = list(range(len(ch_names)))

    if to_keep is not None:
        for i,ch in enumerate(to_keep):
            if isinstance(ch,str):
                ch_idx = ch_names.index(ch.lower())
                to_keep[i] = ch_idx
        load_channels = list(to_keep) # copy list compatible with py2.7
    elif to_drop is not None:
        for i,ch in enumerate(to_drop):
            if isinstance(ch,str):
                ch_idx = ch_names.index(ch.lower())
                to_drop[i] = ch_idx
        to_drop = [len(ch_nrs)+ch if ch<0 else ch for ch in to_drop]

        [ch_nrs.remove(ch) for ch in to_drop]
        load_channels = list(ch_nrs)
    else:
        raise ValueError

    signals, signal_headers, header = read_edf(edf_source,
                                               ch_nrs=load_channels,
                                               digital=True, verbose=verbose)

    write_edf(edf_target, signals, signal_headers, header, digital=True)
    return edf_target


def anonymize_edf(edf_file, new_file=None,
                  to_remove=['patientname', 'birthdate'],
                  new_values=['xxx', ''], verify=False, verbose=False):
    """Anonymize an EDF file by replacing values of header fields.

    This function can be used to overwrite all header information that is
    patient specific, for example birthdate and patientname. All header fields
    can be overwritten this way (i.e., all header.keys() given
    _, _, header = read_edf(edf_file, digital=True)).

    Parameters
    ----------
    edf_file : str
         Filename of an EDF/BDF.
    new_file : str | None
         The filename of the anonymized file. If None, the input filename
         appended with '_anonymized' is used. Defaults to None.
    to_remove : list of str
        List of attributes to overwrite in the `edf_file`. Defaults to
        ['patientname', 'birthdate'].
    new_values : list of str
        List of values used for overwriting the attributes specified in
        `to_remove`. Each item in `to_remove` must have a corresponding item
        in `new_values`. Defaults to ['xxx', ''].
    verify : bool
        Compare `edf_file` and `new_file` for equality (i.e., double check that
        values are same). Defaults to False
    verbose : bool, optional
        print progress or not. The default is False.

    Returns
    -------
    bool
        True if successful, or if `verify` is False. Raises an error otherwise.

    """
    if not len(to_remove) == len(new_values):
        raise AssertionError('Each `to_remove` must have one `new_value`')

    if new_file is None:
        file, ext = os.path.splitext(edf_file)
        new_file = file + '_anonymized' + ext

    signals, signal_headers, header = read_edf(edf_file, digital=True,
                                               verbose=verbose)

    for new_val, attr in zip(new_values, to_remove):
        header[attr] = new_val

    write_edf(new_file, signals, signal_headers, header, digital=True)
    if verify:
        compare_edf(edf_file, new_file, verbose=verbose)
    return True


def rename_channels(edf_file, mapping, new_file=None, verbose=False):
    """
    A convenience function to rename channels in an EDF file.

    Parameters
    ----------
    edf_file : str
        an string pointing to an edf file.
    mapping : dict
         a dictionary with channel mappings as key:value.
         eg: {'M1-O2':'A1-O2'}
    new_file : str, optional
        the new filename. If None will be edf_file + '_renamed'
        The default is None.
    verbose : bool, optional
        print progress or not. The default is False.

    Returns
    -------
    bool
        True if successful, False if failed.

    """
    header = read_edf_header(edf_file)
    channels = header['channels']
    if new_file is None:
        file, ext = os.path.splitext(edf_file)
        new_file = file + '_renamed' + ext

    signal_headers = []
    signals = []
    for ch_nr in tqdm(range(len(channels)), disable=not verbose):
        signal, signal_header, _ = read_edf(edf_file, digital=True,
                                            ch_nrs=ch_nr, verbose=verbose)
        ch = signal_header[0]['label']
        if ch in mapping :
            if verbose: print(f'{ch} to {mapping[ch]}')
            ch = mapping[ch]
            signal_header[0]['label']=ch
        else:
            if verbose: print(f'no mapping for {ch}, leave as it is')
        signal_headers.append(signal_header[0])
        signals.append(signal.squeeze())

    return write_edf(new_file, signals, signal_headers, header, digital=True)


def change_polarity(edf_file, channels, new_file=None, verify=True,
                    verbose=False):
    """
    Change polarity of certain channels

    Parameters
    ----------
    edf_file : str
        from which file to change polarity.
    channels : list of int
        the indices of the channels.
    new_file : str, optional
        where to save the edf with inverted channels. The default is None.
    verify : bool, optional
        whether to verify the two edfs for similarity. The default is True.
    verbose : str, optional
        print progress or not. The default is True.

    Returns
    -------
    bool
        True if success.

    """

    if new_file is None:
        new_file = os.path.splitext(edf_file)[0] + '.edf'

    if isinstance(channels, str): channels=[channels]
    channels = [c.lower() for c in channels]

    signals, signal_headers, header = read_edf(edf_file, digital=True,
                                               verbose=verbose)
    for i,sig in enumerate(signals):
        label = signal_headers[i]['label'].lower()
        if label in channels:
            if verbose: print(f'inverting {label}')
            signals[i] = -sig
    write_edf(new_file, signals, signal_headers, header,
              digital=True, correct=False, verbose=verbose)
    if verify: compare_edf(edf_file, new_file)
    return True



import os

import numpy as np
import pyedflib


def test_generator():
    """
    Get an sample EDF-file

    Parameters
    ----------
    None

    Returns
    -------
    f : EdfReader object
       object containing the handle to the file

    Examples
    --------
    >>> import pyedflib.data
    >>> f = pyedflib.data.test_generator()
    >>> f.signals_in_file == 11
    True
    >>> f._close()
    >>> del f

    """
    fname = get_generator_filename()
    f = pyedflib.EdfReader(fname)
    return f


def get_generator_filename():
    return os.path.join(os.path.dirname(__file__), 'test_generator.edf')



#!/usr/bin/env python

import numpy as np

import pyedflib


from stacklineplot import stackplot


if __name__ == '__main__':
    f = pyedflib.data.test_generator()
    n = f.signals_in_file
    signal_labels = f.getSignalLabels()
    n_min = f.getNSamples()[0]
    sigbufs = [np.zeros(f.getNSamples()[i]) for i in np.arange(n)]
    for i in np.arange(n):
        sigbufs[i] = f.readSignal(i)
        if n_min < len(sigbufs[i]):
            n_min = len(sigbufs[i])
    f._close()
    del f

    n_plot = np.min((n_min, 2000))
    sigbufs_plot = np.zeros((n, n_plot))
    for i in np.arange(n):
        sigbufs_plot[i,:] = sigbufs[i][:n_plot]

    stackplot(sigbufs_plot[:, :n_plot], ylabels=signal_labels)



# Copyright (c) 2015-2023 Holger Nahrstaedt, Simon Kern
# Copyright (c) 2011, 2015, Chris Lee-Messer
# See LICENSE for license details.

__doc__ = """Cython wrapper for low-level C edflib implementation."""
__all__ = ['lib_version', 'CyEdfReader', 'set_patientcode', 'set_starttime_subsecond',
           'write_annotation_latin1', 'write_annotation_utf8', 'set_technician', 'EdfAnnotation',
           'get_annotation', 'read_int_samples', 'blockwrite_digital_samples', 'blockwrite_physical_samples',
           'set_recording_additional', 'write_physical_samples' ,'set_patientname', 'set_physical_minimum',
           'read_physical_samples', 'close_file', 'set_physical_maximum', 'open_file_writeonly',
           'set_patient_additional', 'set_digital_maximum', 'set_birthdate', 'set_digital_minimum',
           'write_digital_samples', 'set_equipment', 'set_samples_per_record','set_admincode', 'set_label',
           'tell', 'rewind', 'set_sex', 'set_gender', 'set_physical_dimension', 'set_transducer',
           'set_prefilter', 'seek', 'set_startdatetime' ,'set_datarecord_duration',
           'set_number_of_annotation_signals', 'open_errors', 'FILETYPE_EDFPLUS',
           'FILETYPE_EDF','FILETYPE_BDF','FILETYPE_BDFPLUS', 'write_errors', 'get_number_of_open_files',
           'get_handle', 'is_file_used', 'blockwrite_digital_short_samples', 'write_digital_short_samples']


#from c_edf cimport *
import locale
import os
import warnings
cimport c_edf
cimport cpython
import numpy as np
cimport numpy as np
from datetime import datetime, date
from cpython.version cimport PY_MAJOR_VERSION
include "edf.pxi"

open_errors = {
    EDFLIB_MALLOC_ERROR : "malloc error",
    EDFLIB_NO_SUCH_FILE_OR_DIRECTORY   : "can not open file, no such file or directory",
    EDFLIB_FILE_CONTAINS_FORMAT_ERRORS : "the file is not EDF(+) or BDF(+) compliant (it contains format errors)",
    EDFLIB_MAXFILES_REACHED            : "to many files opened",
    EDFLIB_FILE_READ_ERROR             : "a read error occurred",
    EDFLIB_FILE_ALREADY_OPENED         : "file has already been opened",
    EDFLIB_FILETYPE_ERROR              : "Wrong file type",
    EDFLIB_FILE_WRITE_ERROR            : "a write error occurred",
    EDFLIB_NUMBER_OF_SIGNALS_INVALID   : "The number of signals is invalid",
    EDFLIB_FILE_IS_DISCONTINUOUS       : "The file is discontinuous and cannot be read",
    EDFLIB_INVALID_READ_ANNOTS_VALUE   : "an annotation value could not be read",
    EDFLIB_FILE_ERRORS_STARTDATE      : "the file is not EDF(+) or BDF(+) compliant, the startdate is incorrect, it might contain incorrect characters, such as ':' instead of '.'",
    EDFLIB_FILE_ERRORS_STARTTIME      : "the file is not EDF(+) or BDF(+) compliant, the starttime is incorrect, it might contain incorrect characters, such as ':' instead of '.'",
    EDFLIB_FILE_ERRORS_NUMBER_SIGNALS : "the file is not EDF(+) or BDF(+) compliant (number of signals)",
    EDFLIB_FILE_ERRORS_BYTES_HEADER   : "the file is not EDF(+) or BDF(+) compliant (Bytes Header)",
    EDFLIB_FILE_ERRORS_RESERVED_FIELD : "the file is not EDF(+) or BDF(+) compliant (Reserved field)",
    EDFLIB_FILE_ERRORS_NUMBER_DATARECORDS : "the file is not EDF(+) or BDF(+) compliant (Number of Datarecords)",
    EDFLIB_FILE_ERRORS_DURATION : "the file is not EDF(+) or BDF(+) compliant (Duration)",
    EDFLIB_FILE_ERRORS_LABEL : "the file is not EDF(+) or BDF(+) compliant the label is incorrect",
    EDFLIB_FILE_ERRORS_TRANSDUCER : "the file is not EDF(+) or BDF(+) compliant the transducer is incorrect",
    EDFLIB_FILE_ERRORS_PHYS_DIMENSION : "the file is not EDF(+) or BDF(+) compliant (Physical Dimension)",
    EDFLIB_FILE_ERRORS_PHYS_MAX : "the file is not EDF(+) or BDF(+) compliant (Physical Maximum)",
    EDFLIB_FILE_ERRORS_PHYS_MIN : "the file is not EDF(+) or BDF(+) compliant (Physical Minimum)",
    EDFLIB_FILE_ERRORS_DIG_MAX : "the file is not EDF(+) or BDF(+) compliant (Digital Maximum)",
    EDFLIB_FILE_ERRORS_DIG_MIN : "the file is not EDF(+) or BDF(+) compliant (Digital Minimum)",
    EDFLIB_FILE_ERRORS_PREFILTER : "the file is not EDF(+) or BDF(+) compliant (Prefilter)",
    EDFLIB_FILE_ERRORS_SAMPLES_DATARECORD : "the file is not EDF(+) or BDF(+) compliant (Sample in Datarecord)",
    EDFLIB_FILE_ERRORS_FILESIZE : "the file is not EDF(+) or BDF(+) compliant (Filesize)",
    EDFLIB_FILE_ERRORS_RECORDINGFIELD : "the file is not EDF(+) or BDF(+) compliant (EDF+ Recordingfield)",
    EDFLIB_FILE_ERRORS_PATIENTNAME : "the file is not EDF(+) or BDF(+) compliant (EDF+ Patientname)",
    'default' : "unknown error"
    }

write_errors = {
    EDFLIB_MALLOC_ERROR                 : "malloc error",
    EDFLIB_NO_SUCH_FILE_OR_DIRECTORY    : "can not open file, no such file or directory",
    EDFLIB_MAXFILES_REACHED             : "to many files opened",
    EDFLIB_FILE_ALREADY_OPENED          : "file has already been opened",
    EDFLIB_FILETYPE_ERROR               : "Wrong file type",
    EDFLIB_FILE_WRITE_ERROR             : "a write error occurred",
    EDFLIB_NUMBER_OF_SIGNALS_INVALID    : "The number of signals is invalid",
    EDFLIB_NO_SIGNALS                   : "no signals to write",
    EDFLIB_TOO_MANY_SIGNALS             : "too many signals",
    EDFLIB_NO_SAMPLES_IN_RECORD         : "no samples in record",
    EDFLIB_DIGMIN_IS_DIGMAX             : "digmin is equal to digmax",
    EDFLIB_DIGMAX_LOWER_THAN_DIGMIN     : "digmax is lower than digmin",
    EDFLIB_PHYSMIN_IS_PHYSMAX           : "physmin is physmax",
    'default' : "unknown error"
}

# constants are redeclared here so we can access them from Python
FILETYPE_EDF = EDFLIB_FILETYPE_EDF
FILETYPE_EDFPLUS = EDFLIB_FILETYPE_EDFPLUS
FILETYPE_BDF = EDFLIB_FILETYPE_BDF
FILETYPE_BDFPLUS = EDFLIB_FILETYPE_BDFPLUS


def contains_unicode(string):
        try:
            string.encode('ascii')
            return False
        except:
            return True

def get_short_path_name(long_name):
    """
    Gets the short path name of a given long path.
    https://stackoverflow.com/a/23598461/200291
    """
    import ctypes
    from ctypes import wintypes
    _GetShortPathNameW = ctypes.windll.kernel32.GetShortPathNameW
    _GetShortPathNameW.argtypes = [wintypes.LPCWSTR, wintypes.LPWSTR, wintypes.DWORD]
    _GetShortPathNameW.restype = wintypes.DWORD
    output_buf_size = 0
    while True:
        output_buf = ctypes.create_unicode_buffer(output_buf_size)
        needed = _GetShortPathNameW(long_name, output_buf, output_buf_size)
        if output_buf_size >= needed:
            return output_buf.value
        else:
            output_buf_size = needed

def lib_version():
    return c_edf.edflib_version()

cdef class CyEdfReader:
    """
    This provides a simple interface to read EDF, EDF+, and probably is ok with
    BDF and BDF+ files
    Note that edflib.c is encapsulated so there is no direct access to the file
    from here unless I add a raw interface or something

    EDF/BDF+ files are arranged into N signals sampled at rate Fs. The data is actually stored in chunks called    "datarecords" which have a file specific size.

    A typical way to use this to read an EEG file would be to choose a certain
    number of seconds per page to display. Then figureout how many data records
    that is. Then read in that many data records at a time. Transform the data as
    needed according the montage and filter settings, then display the data.

    """


    cdef c_edf.edf_hdr_struct hdr
    cdef size_t nsamples_per_record
    #I think it is ok not to do this in __cinit__(*,**)
    def __init__(self, file_name, annotations_mode=EDFLIB_READ_ALL_ANNOTATIONS, check_file_size=EDFLIB_CHECK_FILE_SIZE):
        """
        EdfReader(file_name, annotations_mode, check_file_size)
        """
        self.hdr.handle = -1
        try:
            self.open(file_name, annotations_mode=annotations_mode, check_file_size=check_file_size)
        except FileNotFoundError as e:
            # if files contain Unicode on Windows, and the locale is set incorrectly
            # there can be errors when creating the file.
            # in this case, we can use a workaround and work on the file
            # using short file names (DOS style)
            exists = os.path.isfile(file_name)
            is_windows = os.name == 'nt'
            if exists and is_windows and contains_unicode(file_name):
                # work-around to at least make Unicode files readable at all
                warnings.warn('the filename {} contains Unicode, but Windows does not fully support this. ' \
                              'Please consider changing your locale to support UTF8. Attempting to '
                              'load file via workaround (https://github.com/holgern/pyedflib/pull/100) '.format(file_name))
                file_name = get_short_path_name(file_name)
                self.open(file_name, annotations_mode=annotations_mode, check_file_size=check_file_size)
            elif exists:
                raise OSError(123, 'File {} was found but can\'t be accessed. ' \
                              'Make sure it contains no special characters ' \
                              'or change your locale to use UTF8.'.format(file_name), None, 123)
            else:
                raise e


    def __dealloc__(self):
        if self.hdr.handle >= 0:
            c_edf.edfclose_file(self.hdr.handle)
            self.hdr.handle = -1

    def check_open_ok(self,result):
        if result == 0:
            return True
        else:
            msg = open_errors[self.hdr.filetype]
            if 'no such file or directory' in msg:
                raise FileNotFoundError, '{}: {}'.format(self.file_name, msg)
            raise OSError, '{}: {}'.format(self.file_name, msg)
            # return False

    def make_buffer(self):
        """
        utility function to make a buffer that can hold a single datarecord. This will
        hold the physical samples for a single data record as a numpy tensor.

        -  might extend to provide for N datarecord size

        """
        print ("self.hdr.datarecords_in_file", self.hdr.datarecords_in_file)
        tmp =0
        for ii in range(self.signals_in_file):
            tmp += self.samples_in_datarecord(ii)
        self.nsamples_per_record = tmp
        dbuffer = np.zeros(tmp, dtype='float64') # will get physical samples, not the original digital samples
        return dbuffer

    def open(self, file_name, annotations_mode=EDFLIB_READ_ALL_ANNOTATIONS, check_file_size=EDFLIB_CHECK_FILE_SIZE):
        """
        open(file_name, annotations_mode, check_file_size)
        """
        file_name_str = file_name.encode('utf_8','strict')
        result = c_edf.edfopen_file_readonly(file_name_str, &self.hdr, annotations_mode, check_file_size)

        self.file_name = file_name

        return self.check_open_ok(result)

    def read_annotation(self):
        cdef c_edf.edf_annotation_struct annot
        annotlist = [['','',''] for x in range(self.annotations_in_file)]
        for ii in range(self.annotations_in_file):
            c_edf.edf_get_annotation(self.hdr.handle, ii, &(annot))
            #get_annotation(self.hdr.handle, ii, &annotation)
            annotlist[ii][0] = annot.onset
            annotlist[ii][1] = annot.duration
            annotlist[ii][2] = annot.annotation
        return annotlist



    property handle:
        "edflib internal int handle"
        def __get__(self):
            return self.hdr.handle

    property datarecords_in_file:
        "number of data records"
        def __get__(self):
            return self.hdr.datarecords_in_file

    property signals_in_file:
        def __get__(self):
            return self.hdr.edfsignals

    property file_duration:
        "file duration in seconds"
        def __get__(self):
            # duration is saved in resolution of 100 ns
            # therefore multiplying with EDFLIB_TIME_DIMENSION
            return self.hdr.file_duration/EDFLIB_TIME_DIMENSION

    property filetype:
        def __get__(self):
            return self.hdr.filetype

    property patient:
        "patient info (legacy EDF format)"
        def __get__(self):
            return self.hdr.patient.rstrip()

    property recording:
        "recording info (legacy EDF format)"
        def __get__(self):
            return self.hdr.recording.rstrip()

    property datarecord_duration:
        "datarecord duration in seconds (as a double)"
        def __get__(self):
            # duration is saved in resolution of 100 ns
            # therefore multiplying with EDFLIB_TIME_DIMENSION
            return (<double>self.hdr.datarecord_duration) / EDFLIB_TIME_DIMENSION

    property annotations_in_file:
        def __get__(self):
            return self.hdr.annotations_in_file

    property patientcode:
        def __get__(self):
            return self.hdr.patientcode

    property sex:
        def __get__(self):
            return self.hdr.gender

    property gender:
        def __get__(self):
            warnings.warn("Variable 'gender' is deprecated, use 'sex' instead.", DeprecationWarning, stacklevel=2)
            return self.hdr.gender

    property birthdate:
        def __get__(self):
            return self.hdr.birthdate

    property patientname:
        def __get__(self):
            return self.hdr.patient_name

    property patient_additional:
        def __get__(self):
            return self.hdr.patient_additional

    property startdate_year:
        def __get__(self):
            return self.hdr.startdate_year

    property startdate_month:
        def __get__(self):
            return self.hdr.startdate_month

    property startdate_day:
        def __get__(self):
            return self.hdr.startdate_day

    property starttime_hour:
        def __get__(self):
            return self.hdr.starttime_hour

    property starttime_minute:
        def __get__(self):
            return self.hdr.starttime_minute

    property starttime_second:
        def __get__(self):
            return self.hdr.starttime_second

    property starttime_subsecond:
        def __get__(self):
            return self.hdr.starttime_subsecond

    property admincode:
        def __get__(self):
            return self.hdr.admincode

    property technician:
        def __get__(self):
            return self.hdr.technician

    property equipment:
        def __get__(self):
            return self.hdr.equipment

    property recording_additional:
        def __get__(self):
            return self.hdr.recording_additional

    # signal parameters
    def signal_label(self, channel):
        return self.hdr.signalparam[channel].label

    def samples_in_file(self,channel):
        return self.hdr.signalparam[channel].smp_in_file

    def samples_in_datarecord(self, channel):
        return self.hdr.signalparam[channel].smp_in_datarecord

    def physical_dimension(self, channel):
        return self.hdr.signalparam[channel].physdimension

    def physical_max(self, channel):
        return self.hdr.signalparam[channel].phys_max

    def physical_min(self, channel):
        return self.hdr.signalparam[channel].phys_min

    def digital_max(self, channel):
        return self.hdr.signalparam[channel].dig_max

    def digital_min(self, channel):
        return self.hdr.signalparam[channel].dig_min

    def prefilter(self, channel):
        return self.hdr.signalparam[channel].prefilter

    def transducer(self, channel):
        return self.hdr.signalparam[channel].transducer

    def samplefrequency(self, channel):
        smp_per_record = <double>self.smp_per_record(channel)
        record_duration = self.datarecord_duration
        return smp_per_record / record_duration


    def smp_per_record(self, channel):
        return <int>self.hdr.signalparam[channel].smp_in_datarecord

    # def _tryoffset0(self):
    #     """
    #     fooling around to find offset in file to allow shortcut mmap interface
    #     """
    #     # cdef long offset = self.hdr.hdrsize  # from edflib.c read_physical_samples()
    #     print "trying to find data offset in file"
    #     nrecords = self.hdr.datarecords_in_file
    #     print "nrecords in file:", nrecords
    #     return 1,2
    #     # return offset, nrecords
    #     # print "offset via edftell:",  edftell(self.hdr.handle, 0)


    def _close(self):   # should not be closed from python
        if self.hdr.handle >= 0:
            c_edf.edfclose_file(self.hdr.handle)
        self.hdr.handle = -1

    def read_digital_signal(self, signalnum, start, n, np.ndarray[np.int32_t, ndim=1] sigbuf):
        """
        read_digital_signal(self, signalnum, start, n, np.ndarray[np.int32_t, ndim=1] sigbuf
        read @n number of samples from signal number @signum starting at @start
        into numpy int32 array @sigbuf sigbuf must be at least n long
        """
        c_edf.edfseek(self.hdr.handle, signalnum, start, EDFSEEK_SET)
        readn = read_int_samples(self.hdr.handle, signalnum, n, sigbuf)
        if readn != n:
            print("read %d, less than %d requested!!!" % (readn, n))

    def readsignal(self, signalnum, start, n, np.ndarray[np.float64_t, ndim=1] sigbuf):

        """read @n number of samples from signal number @signum starting at
        @start into numpy float64 array @sigbuf sigbuf must be at least n long
        """

        c_edf.edfseek(self.hdr.handle, signalnum, start, EDFSEEK_SET)
        readn = c_edf.edfread_physical_samples(self.hdr.handle, signalnum, n, <double*>sigbuf.data)
        # print "read %d samples" % readn
        if readn != n:
            print ("read %d, less than %d requested!!!" % (readn, n))

    def load_datarecord(self, np.ndarray[np.float64_t, ndim=1] db, n=0):
        cdef size_t offset =0

        if n < self.hdr.datarecords_in_file:
            for ii in range(self.signals_in_file):
                c_edf.edfseek(self.hdr.handle, ii, n*self.samples_in_datarecord(ii), EDFSEEK_SET) # just a guess
                readn = c_edf.edfread_physical_samples(self.hdr.handle, ii, self.samples_in_datarecord(ii),
                                                 (<double*>db.data)+offset)
                print ("readn this many samples", readn)
                offset += self.samples_in_datarecord(ii)


###############################
# low level functions


def set_patientcode(int handle, char *patientcode):
    # check if rw?
    return c_edf.edf_set_patientcode(handle, patientcode)

cpdef int write_annotation_latin1(int handle, long long onset, long long duration, char *description):
        return c_edf.edfwrite_annotation_latin1(handle, onset, duration, description)

cpdef int write_annotation_utf8(int handle, long long onset, long long duration, char *description):
        return c_edf.edfwrite_annotation_utf8(handle, onset, duration, description)

cpdef int set_technician(int handle, char *technician):
    return c_edf.edf_set_technician(handle, technician)

cdef class EdfAnnotation:
    cdef c_edf.edf_annotation_struct annotation


cpdef int get_annotation(int handle, int n, EdfAnnotation edf_annotation):
    return c_edf.edf_get_annotation(handle, n, &(edf_annotation.annotation))

# need to use npbuffers
cpdef read_int_samples(int handle, int edfsignal, int n,
                         np.ndarray[np.int32_t,ndim=1] buf):
    """
    reads n samples from edfsignal, starting from the current sample position indicator, into buf (edfsignal starts at 0)
    the values are the "raw" digital values
    bufsize should be equal to or bigger than sizeof(int[n])
    the sample position indicator will be increased with the amount of samples read
    returns the amount of samples read (this can be less than n or zero!)
    or -1 in case of an error


    ToDO!!!
    assert that these are stored as EDF/EDF+ files with int16 sized samples
    returns how many were actually read
    doesn't currently check that buf can hold all the data
    """
    return c_edf.edfread_digital_samples(handle, edfsignal, n,<int*>buf.data)

cpdef int blockwrite_digital_samples(int handle, np.ndarray[np.int32_t,ndim=1] buf):
    return c_edf.edf_blockwrite_digital_samples(handle, <int*>buf.data)

cpdef int blockwrite_digital_short_samples(int handle, np.ndarray[np.int16_t,ndim=1] buf):
    return c_edf.edf_blockwrite_digital_short_samples(handle, <short*>buf.data)

cpdef int blockwrite_physical_samples(int handle, np.ndarray[np.float64_t,ndim=1] buf):
    return c_edf.edf_blockwrite_physical_samples(handle, <double*>buf.data)

cpdef int set_recording_additional(int handle, char *recording_additional):
    return c_edf.edf_set_recording_additional(handle,recording_additional)

cpdef int write_digital_short_samples(int handle, np.ndarray[np.int16_t] buf):
    return c_edf.edfwrite_digital_short_samples(handle, <short *>buf.data)

cpdef int write_physical_samples(int handle, np.ndarray[np.float64_t] buf):
    return c_edf.edfwrite_physical_samples(handle, <double *>buf.data)

cpdef int set_patientname(int handle, char *name):
    return c_edf.edf_set_patientname(handle, name)

cpdef int set_physical_minimum(int handle, int edfsignal, double phys_min):
    return c_edf.edf_set_physical_minimum(handle, edfsignal, phys_min)

cpdef int read_physical_samples(int handle, int edfsignal, int n,
                                np.ndarray[np.float64_t] buf):
    return c_edf.edfread_physical_samples(handle, edfsignal, n, <double *>buf.data)

def close_file(handle):
    return c_edf.edfclose_file(handle)

def get_number_of_open_files():
    return c_edf.edflib_get_number_of_open_files()

def get_handle(file_number):
    return c_edf.edflib_get_handle(file_number)

def is_file_used(path):
    path_byte = path.encode('utf_8','strict')
    return c_edf.edflib_is_file_used(path_byte)

# so you can use the same name if defining a python only function
def set_physical_maximum(handle, edfsignal, phys_max):
    return c_edf.edf_set_physical_maximum(handle, edfsignal, phys_max)

def open_file_writeonly(path, filetype, number_of_signals):
    """int edfopen_file_writeonly(char *path, int filetype, int number_of_signals)"""

    if os.name=='nt' and contains_unicode(path):
        default_enc = locale.getdefaultlocale()[1]
        if default_enc is None:
            default_enc = ''
        else:
            default_enc = default_enc.lower()
        using_unicode = 'utf' in default_enc or 'unicode' in default_enc or \
                        '10646' in  default_enc or default_enc=='cp65001'
        # Check if we're on Windows and the file path contains Unicode.
        # If so, use workaround to create file: In Python, create the file,
        # then look up and pass the short file name to the C library
        if not using_unicode:
            warnings.warn('Attempting to write Unicode file {} on Windows. ' \
                          'Consider changing your locale to UTF8.'.format(path))
            with open(path, 'wb'): pass
            path = get_short_path_name(path)

    py_byte_string  = path.encode('utf_8','strict')
    cdef char* path_str = py_byte_string
    return c_edf.edfopen_file_writeonly(path_str, filetype, number_of_signals)

def set_patient_additional(handle, patient_additional):
    """int edf_set_patient_additional(int handle, const char *patient_additional)"""
    return c_edf.edf_set_patient_additional(handle, patient_additional)

def set_digital_maximum(handle, edfsignal, dig_max):
    "int edf_set_digital_maximum(int handle, int edfsignal, int dig_max)"
    return c_edf.edf_set_digital_maximum(handle, edfsignal, dig_max)

# see CyEdfreader() class
# int edfopen_file_readonly(const char *path, struct edf_hdr_struct *edfhdr, int read_annotations)

def set_birthdate(handle, birthdate_year, birthdate_month, birthdate_day):
    """int edf_set_birthdate(int handle, int birthdate_year, int birthdate_month, int birthdate_day)"""
    return c_edf.edf_set_birthdate(handle, birthdate_year,  birthdate_month, birthdate_day)

def set_digital_minimum(handle, edfsignal, dig_min):
    """int edf_set_digital_minimum(int handle, int edfsignal, int dig_min)"""
    return c_edf.edf_set_digital_minimum(handle,  edfsignal, dig_min)

def write_digital_samples(handle, np.ndarray[np.int32_t] buf):
    """write_digital_samples(int handle, np.ndarray[np.int32_t] buf)"""
    return c_edf.edfwrite_digital_samples(handle, <int*>buf.data)

def set_equipment(handle, equipment):
    """int edf_set_equipment(int handle, const char *equipment)"""
    return c_edf.edf_set_equipment(handle, equipment)

def set_samples_per_record(handle, edfsignal, smp_per_record ):
    """
    int set_samples_per_record(int handle, int edfsignal, int smp_per_record )

    sets how many samples are in the record for this signal.
    this is not the sampling frequency (Hz), (which is calculated by
    by smp_per_record/record_duration).
    """
    return c_edf.edf_set_samplefrequency(handle, edfsignal, smp_per_record)

def set_admincode(handle, admincode):
    """int edf_set_admincode(int handle, const char *admincode)"""
    return c_edf.edf_set_admincode(handle, admincode)

def set_label(handle, edfsignal, label):
    """int edf_set_label(int handle, int edfsignal, const char *label)"""
    return c_edf.edf_set_label(handle, edfsignal, label)


#FIXME need to make sure this gives the proper values for large values
def tell(handle, edfsignal):
    """long long edftell(int handle, int edfsignal)"""
    return c_edf.edftell(handle,  edfsignal)

def rewind(handle, edfsignal):
    """void edfrewind(int handle, int edfsignal)"""
    c_edf.edfrewind(handle, edfsignal)

def set_sex(handle, sex):
    """int edf_set_sex(int handle, int sex)"""
    if sex is None: return 0 #don't set sex at all to prevent default 'F'
    return c_edf.edf_set_gender(handle, sex)

def set_gender(handle, gender):
    warnings.warn("Function 'set_gender' is deprecated, use 'set_sex' instead.", DeprecationWarning, stacklevel=2)
    return set_sex(handle, gender)

def set_physical_dimension(handle, edfsignal, phys_dim):
    """int edf_set_physical_dimension(int handle, int edfsignal, const char *phys_dim)"""
    return c_edf.edf_set_physical_dimension(handle, edfsignal, phys_dim)

def set_transducer(handle, edfsignal, transducer):
    """int edf_set_transducer(int handle, int edfsignal, const char *transducer)"""
    return c_edf.edf_set_transducer(handle, edfsignal, transducer)

def set_prefilter(handle, edfsignal, prefilter):
    """int edf_set_prefilter(int handle, int edfsignal, const char*prefilter)"""
    return c_edf.edf_set_prefilter(handle, edfsignal, prefilter)

def seek(handle, edfsignal, offset, whence):
    """long long edfseek(int handle, int edfsignal, long long offset, int whence)"""
    return c_edf.edfseek(handle, edfsignal, offset, whence)

def set_startdatetime(handle, startdate_year, startdate_month, startdate_day,
                                 starttime_hour, starttime_minute, starttime_second):
    """int edf_set_startdatetime(int handle, int startdate_year, int startdate_month, int startdate_day,
                                      int starttime_hour, int starttime_minute, int starttime_second)"""
    return c_edf.edf_set_startdatetime(handle, startdate_year, startdate_month, startdate_day,
                                 starttime_hour, starttime_minute, starttime_second)

def set_starttime_subsecond(handle, subsecond):
    """int edf_set_subsecond_starttime(int handle, int subsecond)"""
    return c_edf.edf_set_subsecond_starttime(handle, subsecond)

def set_datarecord_duration(handle, duration):
    """int edf_set_datarecord_duration(int handle, int duration)

    duration in seconds
    """
    # from the pyedflib documentation:
    # > To avoid rounding errors, the library stores some timevalues in variables
    # > of type long long int. In order not to loose the subsecond precision, all
    # > timevalues have been multiplied by 10000000. This will limit the
    # > timeresolution to 100 nanoSeconds. To calculate the amount of seconds,
    # > divide the timevalue by 10000000 or use the macro EDFLIB_TIME_DIMENSION
    # > which is declared in edflib.h.
    #
    # therefore, we divide by 100, and edflib internally multiplies by 100.
    # we could also change it in the C file, but better to leave that as is.
    duration *= EDFLIB_TIME_DIMENSION/100
    return c_edf.edf_set_datarecord_duration(handle, int(duration))

def set_number_of_annotation_signals(handle, annot_signals):
    """int edf_set_number_of_annotation_signals(int handle, int annot_signals)"""
    return c_edf.edf_set_number_of_annotation_signals(handle, annot_signals)




# THIS FILE IS GENERATED FROM pyedflib SETUP.PY
short_version = '0.1.34'
version = '0.1.34'
full_version = '0.1.34'
git_revision = '1af9ced84921b304fecdc994e7e8b9a4c1674245'
release = True

if not release:
    version = full_version



#!/usr/bin/env python

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import time
import pyedflib
from stacklineplot import stackplot_t

def animate(frame):
    global offset
    for i in np.arange(n):
        sigbufs_plot[i,:] = sigbufs[i][offset:n_plot + offset]
    ax1.clear()
    stackplot_t(np.transpose(sigbufs_plot[:, :n_plot]), ylabels=signal_labels, ax=ax1)
    offset += dt

if __name__ == '__main__':
    f = pyedflib.data.test_generator()
    n = f.signals_in_file
    signal_labels = f.getSignalLabels()
    n_min = f.getNSamples()[0]
    sigbufs = [np.zeros(f.getNSamples()[i]) for i in np.arange(n)]
    for i in np.arange(n):
        sigbufs[i] = f.readSignal(i)
        if n_min < len(sigbufs[i]):
            n_min = len(sigbufs[i])

    duration = f.getFileDuration()
    f._close()
    del f
    dt = int(duration/5)
    fig = plt.figure()
    ax1 = fig.add_subplot(1,1,1)
    n_plot = np.min((n_min, 2000))
    sigbufs_plot = np.zeros((n, n_plot))
    offset = 0
    ani = animation.FuncAnimation(fig, animate, interval=200)
    plt.show()



# Copyright (c) 2019 - 2020 Simon Kern
# Copyright (c) 2015 Holger Nahrstaedt

import os
import gc
import numpy as np
from datetime import datetime
# from numpy.testing import (assert_raises, run_module_suite,
#                            assert_equal, assert_allclose, assert_almost_equal)
import unittest
import pyedflib
from pyedflib.edfwriter import EdfWriter
from pyedflib.edfreader import EdfReader


class TestEdfReader(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # data_dir = os.path.join(os.getcwd(), 'data')
        data_dir = os.path.join(os.path.dirname(__file__), 'data')
        cls.edf_data_file = os.path.join(data_dir, 'test_generator.edf')
        cls.bdf_data_file = os.path.join(data_dir, 'test_generator.bdf')
        cls.bdf_data_file_datarec_2 = os.path.join(data_dir, 'test_generator_datarec_generator_2.bdf')
        cls.bdf_data_file_datarec_0_5 = os.path.join(data_dir, 'test_generator_datarec_generator_0_5.bdf')
        cls.bdf_broken_file = os.path.join(data_dir, 'tmp_broken_file.bdf')
        cls.bdf_accented_file = os.path.join(data_dir, 'tmp_file_áä\'üöß.bdf')
        cls.edf_subsecond = os.path.join(data_dir, 'test_subsecond.edf')
        cls.tmp_edf_file = os.path.join(data_dir, 'test_tmp_file.edf')
        cls.edf_legacy = os.path.join(data_dir, 'test_legacy.edf')

    @classmethod
    def tearDownClass(cls):
        data_dir = os.path.join(os.path.dirname(__file__), 'data')
        tmpfiles = [f for f in os.listdir(data_dir) if f.startswith('tmp')]
        for file in tmpfiles:
            try:
                os.remove(os.path.join(data_dir, file))
            except Exception as e:
                print(e)

    def tearDown(self):
        # small hack to close handles in case of tests throwing an exception
        for obj in gc.get_objects():
            if isinstance(obj, (EdfWriter, EdfReader)):
                obj.close()
                del obj

    def test_EdfReader(self):
        try:
            f = pyedflib.EdfReader(self.edf_data_file)
        except OSError:
            print('cannot open', self.edf_data_file)
            return

        ann_index, ann_duration, ann_text = f.readAnnotations()
        np.testing.assert_almost_equal(ann_index[0], 0)
        np.testing.assert_almost_equal(ann_index[1], 600)

        np.testing.assert_equal(f.signals_in_file, 11)
        np.testing.assert_equal(f.datarecords_in_file, 600)

        for i in np.arange(11):
            np.testing.assert_almost_equal(f.getSampleFrequencies()[i], 200)
            np.testing.assert_equal(f.getNSamples()[i], 120000)
        np.testing.assert_equal(f.handle, 0)
        f.close()
        np.testing.assert_equal(f.handle, -1)

    def test_EdfReader_BDF(self):
        try:
            f = pyedflib.EdfReader(self.bdf_data_file)
        except OSError:
            print('cannot open', self.bdf_data_file)
            return

        datarecords = 30

        np.testing.assert_equal(f.signals_in_file, 5)
        np.testing.assert_equal(f.datarecords_in_file, datarecords)
        np.testing.assert_equal(f.getFileDuration(), datarecords)

        sample_frequencies = [1000, 800, 500, 975, 999]

        for i in np.arange(f.signals_in_file):
            np.testing.assert_almost_equal(f.getSampleFrequencies()[i], sample_frequencies[i])
            np.testing.assert_almost_equal(f.getSampleFrequency(i), sample_frequencies[i])
            np.testing.assert_equal(f.getNSamples()[i], int(sample_frequencies[i] * datarecords))
            np.testing.assert_almost_equal(f.getSignalHeader(i)["sample_frequency"], sample_frequencies[i])
            np.testing.assert_almost_equal(f.getSignalHeaders()[i]["sample_frequency"], sample_frequencies[i])

        np.testing.assert_equal(f.handle, 0)
        f.close()
        np.testing.assert_equal(f.handle, -1)

    def test_EdfReader_BDF_datarec_0_5(self):
        # try:
        f = pyedflib.EdfReader(self.bdf_data_file_datarec_0_5)
        # except OSError:
        #     print('cannot open', self.bdf_data_file_datarec_0_5)
        #     return

        datarecords = 30
        datarecord_duration = 0.5

        np.testing.assert_equal(f.signals_in_file, 5)
        np.testing.assert_equal(f.datarecord_duration, datarecord_duration)
        np.testing.assert_equal(f.datarecords_in_file, datarecords / datarecord_duration)
        np.testing.assert_equal(f.getFileDuration(), datarecords)

        sample_frequencies = [2000, 1600, 1000, 1950, 1998]

        for i in np.arange(f.signals_in_file):
            np.testing.assert_almost_equal(f.getSampleFrequencies()[i], sample_frequencies[i])
            np.testing.assert_almost_equal(f.getSampleFrequency(i), sample_frequencies[i])
            np.testing.assert_equal(f.getNSamples()[i], int(sample_frequencies[i] * datarecords))
            np.testing.assert_almost_equal(f.getSignalHeader(i)["sample_frequency"], sample_frequencies[i])
            np.testing.assert_almost_equal(f.getSignalHeaders()[i]["sample_frequency"], sample_frequencies[i])
        np.testing.assert_equal(f.handle, 0)
        f.close()
        np.testing.assert_equal(f.handle, -1)

    def test_EdfReader_BDF_datarec_2(self):
        # try:
        f = pyedflib.EdfReader(self.bdf_data_file_datarec_2)
        # except OSError:
        #     print('cannot open', self.bdf_data_file_datarec_2)
        #     return

        datarecords = 30
        datarecord_duration = 2

        np.testing.assert_equal(f.signals_in_file, 5)
        np.testing.assert_equal(f.datarecord_duration, datarecord_duration)
        np.testing.assert_equal(f.datarecords_in_file, datarecords / datarecord_duration)
        np.testing.assert_equal(f.getFileDuration(), datarecords)

        sample_frequencies = [500, 400, 250, 487.5, 499.5]

        for i in np.arange(f.signals_in_file):
            np.testing.assert_almost_equal(f.getSampleFrequencies()[i], sample_frequencies[i])
            np.testing.assert_almost_equal(f.getSampleFrequency(i), sample_frequencies[i])
            np.testing.assert_equal(f.getNSamples()[i], int(sample_frequencies[i] * datarecords))
            np.testing.assert_almost_equal(f.getSignalHeader(i)["sample_frequency"], sample_frequencies[i])
            np.testing.assert_almost_equal(f.getSignalHeaders()[i]["sample_frequency"], sample_frequencies[i])
        np.testing.assert_equal(f.handle, 0)
        f.close()
        np.testing.assert_equal(f.handle, -1)

    def test_indexerrors_thrown(self):
        try:
            f = pyedflib.EdfReader(self.edf_data_file)
        except OSError:
            print('cannot open', self.edf_data_file)
            return

        funcs = ['getSampleFrequency', 'getLabel', 'getPrefilter', 'getPhysicalMaximum',
                 'getPhysicalMinimum', 'getDigitalMaximum', 'getDigitalMinimum',
                 'getTransducer', 'getPhysicalDimension', 'readSignal']
        for i in range(10):
            for func in funcs:
                getattr(f, func)(i)

        for i in [-1, 11]:
            for func in funcs:
                with self.assertRaises(IndexError, msg=f"f.{func}({i})"):
                    getattr(f, func)(i)

        f.close()

    def test_EdfReader_headerInfos(self):
        try:
            f = pyedflib.EdfReader(self.edf_data_file)
        except OSError:
            print('cannot open', self.edf_data_file)
            return
        datetimeSoll = datetime(2011,4,4,12,57,2)
        np.testing.assert_equal(f.getStartdatetime(),datetimeSoll)
        np.testing.assert_equal(f.getPatientCode(), 'abcxyz99')
        np.testing.assert_equal(f.getPatientName(), 'Hans Muller')
        np.testing.assert_equal(f.getSex(), 'Male')
        np.testing.assert_equal(f.getGender(), 'Male')  # deprecated
        np.testing.assert_equal(f.getBirthdate(), '30 jun 1969')
        np.testing.assert_equal(f.getPatientAdditional(), 'patient')
        np.testing.assert_equal(f.getAdmincode(), 'Dr. X')
        np.testing.assert_equal(f.getTechnician(), 'Mr. Spotty')
        np.testing.assert_equal(f.getEquipment(), 'test generator')
        np.testing.assert_equal(f.getRecordingAdditional(), 'unit test file')
        np.testing.assert_equal(f.getFileDuration(), 600)
        fileHeader = f.getHeader()
        np.testing.assert_equal(fileHeader["patientname"], 'Hans Muller')
        f.close()

    def test_EdfReader_signalInfos(self):
        try:
            f = pyedflib.EdfReader(self.edf_data_file)
        except OSError:
            print('cannot open', self.edf_data_file)
            return
        np.testing.assert_equal(f.getSignalLabels()[0], 'squarewave')
        np.testing.assert_equal(f.getLabel(0), 'squarewave')
        np.testing.assert_equal(f.getPrefilter(0), 'pre1')
        np.testing.assert_equal(f.getTransducer(0), 'trans1')
        for i in np.arange(1,11):
            np.testing.assert_equal(f.getPhysicalDimension(i), 'uV')
            np.testing.assert_equal(f.getSampleFrequency(i), 200)
            np.testing.assert_equal(f.getSampleFrequencies()[i], 200)

        np.testing.assert_equal(f.getSignalLabels()[1], 'ramp')
        np.testing.assert_equal(f.getSignalLabels()[2], 'pulse')
        np.testing.assert_equal(f.getSignalLabels()[3], 'noise')
        np.testing.assert_equal(f.getSignalLabels()[4], 'sine 1 Hz')
        np.testing.assert_equal(f.getSignalLabels()[5], 'sine 8 Hz')
        np.testing.assert_equal(f.getSignalLabels()[6], 'sine 8.1777 Hz')
        np.testing.assert_equal(f.getSignalLabels()[7], 'sine 8.5 Hz')
        np.testing.assert_equal(f.getSignalLabels()[8], 'sine 15 Hz')
        np.testing.assert_equal(f.getSignalLabels()[9], 'sine 17 Hz')
        np.testing.assert_equal(f.getSignalLabels()[10], 'sine 50 Hz')
        # testing file info and file_info_log
        f.file_info()
        f.file_info_long()
        f.close()

    def test_EdfReader_subsecond(self):
        f = pyedflib.EdfReader(self.edf_subsecond)
        self.assertEqual(f.getStartdatetime(), datetime(2020, 1, 24, 4, 5, 56, 39453))


    def test_EdfReader_ReadAnnotations(self):
        try:
            f = pyedflib.EdfReader(self.edf_data_file, pyedflib.DO_NOT_READ_ANNOTATIONS)
        except OSError:
            print('cannot open', self.edf_data_file)
            return

        ann_index, ann_duration, ann_text = f.readAnnotations()
        np.testing.assert_equal(ann_index.size, 0)

        del f

        try:
            f = pyedflib.EdfReader(self.edf_data_file, pyedflib.READ_ALL_ANNOTATIONS)
        except OSError:
            print('cannot open', self.edf_data_file)
            return

        ann_index, ann_duration, ann_text = f.readAnnotations()
        np.testing.assert_almost_equal(ann_index[0], 0)
        np.testing.assert_almost_equal(ann_index[1], 600)

        np.testing.assert_equal(f.signals_in_file, 11)
        np.testing.assert_equal(f.datarecords_in_file, 600)

        for i in np.arange(11):
            np.testing.assert_almost_equal(f.getSampleFrequencies()[i], 200)
            np.testing.assert_equal(f.getNSamples()[i], 120000)

        del f

    def test_EdfReader_Check_Size(self):
        sample_frequency = 100
        channel_info = {'label': 'test_label', 'dimension': 'mV', 'sample_rate': 256,
                        'sample_frequency': sample_frequency, 'physical_max': 1.0, 'physical_min': -1.0,
                        'digital_max': 8388607, 'digital_min': -8388608,
                        'prefilter': 'pre1', 'transducer': 'trans1'}
        f = pyedflib.EdfWriter(self.bdf_broken_file, 1, file_type=pyedflib.FILETYPE_BDFPLUS)
        f.setSignalHeader(0,channel_info)
        f.setTechnician('tec1')
        data = np.ones(100) * 0.1
        f.writePhysicalSamples(data)
        f.writePhysicalSamples(data)
        del f

        f = pyedflib.EdfReader(self.bdf_broken_file, pyedflib.READ_ALL_ANNOTATIONS, pyedflib.DO_NOT_CHECK_FILE_SIZE)
        np.testing.assert_equal(f.getTechnician(), 'tec1')

        np.testing.assert_equal(f.getLabel(0), 'test_label')
        np.testing.assert_equal(f.getPhysicalDimension(0), 'mV')
        np.testing.assert_equal(f.getPrefilter(0), 'pre1')
        np.testing.assert_equal(f.getTransducer(0), 'trans1')
        np.testing.assert_equal(f.getSampleFrequency(0), sample_frequency)

        # When both 'sample_rate' and 'sample_frequency' are present, we write
        # the file using the latter, which means that when we read it back,
        # only the 'sample_frequency' value is present.
        expected_signal_header = {**channel_info, 'sample_rate': sample_frequency}
        np.testing.assert_equal(f.getSignalHeader(0), expected_signal_header)
        np.testing.assert_equal(f.getSignalHeaders(), [expected_signal_header])
        del f

    def test_EdfReader_Close_file(self):
        f = pyedflib.EdfReader(self.edf_data_file)
        np.testing.assert_equal(f.getSignalLabels()[0], 'squarewave')

        # Don't close the file but try to reopen it and verify that it fails.
        with np.testing.assert_raises(OSError):
            ff = pyedflib.EdfReader(self.edf_data_file)

        # Now close and verify it can be re-opened/read.
        f.close()

        ff = pyedflib.EdfReader(self.edf_data_file)
        np.testing.assert_equal(ff.getSignalLabels()[0], 'squarewave')
        del f, ff

    def test_BdfReader_Read_accented_file(self):
        sample_frequency = 100
        channel_info = {'label': 'test_label', 'dimension': 'mV', 'sample_rate': 256,
                        'sample_frequency': sample_frequency, 'physical_max': 1.0, 'physical_min': -1.0,
                        'digital_max': 8388607, 'digital_min': -8388608,
                        'prefilter': 'pre1', 'transducer': 'trans1'}
        f = pyedflib.EdfWriter(self.bdf_accented_file, 1,file_type=pyedflib.FILETYPE_BDFPLUS)
        f.setSignalHeader(0,channel_info)
        f.setTechnician('tec1')
        data = np.ones(100) * 0.1
        f.writePhysicalSamples(data)
        f.writePhysicalSamples(data)
        del f

        f = pyedflib.EdfReader(self.bdf_accented_file, pyedflib.READ_ALL_ANNOTATIONS, pyedflib.DO_NOT_CHECK_FILE_SIZE)
        np.testing.assert_equal(f.getTechnician(), 'tec1')
        np.testing.assert_equal(f.getLabel(0), 'test_label')
        np.testing.assert_equal(f.getPhysicalDimension(0), 'mV')
        np.testing.assert_equal(f.getPrefilter(0), 'pre1')
        np.testing.assert_equal(f.getTransducer(0), 'trans1')

        # When both 'sample_rate' and 'sample_frequency' are present, we write
        # the file using the latter, which means that when we read it back,
        # only the 'sample_frequency' value is present.
        expected_signal_header = {**channel_info, 'sample_rate': sample_frequency}
        np.testing.assert_equal(f.getSignalHeader(0), expected_signal_header)
        np.testing.assert_equal(f.getSignalHeaders(), [expected_signal_header])
        del f


    def test_read_incorrect_file(self):

        # this should simply not be found and not trigger the UTF8 warning
        with self.assertRaises(FileNotFoundError):
            f = pyedflib.EdfReader('does_not_exist')

        # this file is corrupted and should be found, but not be read
        # so the Exception should be OSError and not FileNotFoundError
        with self.assertRaises(OSError):
            with open(self.tmp_edf_file, 'wb') as f:
                f.write(b'0123456789')
            f = pyedflib.EdfReader(self.tmp_edf_file)

        # now we create a file that is not EDF/BDF compliant
        # this should raise an OSError and not a FileNotFoundError
        with self.assertRaises(OSError) as cm:
            channel_info = {'label': 'test_label', 'dimension': 'mV', 'sample_rate': 256,
                            'sample_frequency': 100, 'physical_max': 1.0, 'physical_min': -1.0,
                            'digital_max': 8388607, 'digital_min': -8388608,
                            'prefilter': 'pre1', 'transducer': 'trans1'}
            with pyedflib.EdfWriter(self.bdf_accented_file, 1, file_type=pyedflib.FILETYPE_BDFPLUS) as f:
                f.setSignalHeader(0,channel_info)
                f.setTechnician('tec1')
                data = np.ones(100) * 0.1
                f.writePhysicalSamples(data)

            with open(self.bdf_accented_file, 'rb') as f:
                content = bytearray(f.read())
                content[181] = 58
            with open(self.bdf_accented_file, 'wb') as f:
                f.write(content)

            f = pyedflib.EdfReader(self.bdf_accented_file)
        # However, the error before should not be a FileNotFoundError
        assert not isinstance(cm.exception, FileNotFoundError)

    def test_EdfReader_Legacy_Header_Info(self):
        expected_header = {
            # Legacy header fields
            'patient': b'Legacy patient description',
            'recording': b'Legacy recording description',
            # All the rest must be empty
            'technician': b'',
            'recording_additional': b'',
            'patientname': b'',
            'patient_additional': b'',
            'patientcode': b'',
            'equipment': b'',
            'admincode': b'',
            'sex': b'',
            'gender': b'',  # deprecated
            'birthdate': b''
        }

        with pyedflib.EdfReader(self.edf_legacy) as f:
            for attr_name, expected_value in expected_header.items():
                self.assertEqual(getattr(f, attr_name), expected_value)

if __name__ == '__main__':
    # run_module_suite(argv=sys.argv)
    unittest.main()



# Copyright (c) 2019 - 2023 Simon Kern
# Copyright (c) 2015 - 2023 Holger Nahrstaedt
# Copyright (c) 2011, 2015, Chris Lee-Messer
# Copyright (c) 2016-2017 The pyedflib Developers
#                         <https://github.com/holgern/pyedflib>
# See LICENSE for license details.

import numpy as np
import sys
import warnings
from datetime import datetime, date
from ._extensions._pyedflib import FILETYPE_EDFPLUS, FILETYPE_BDFPLUS, FILETYPE_BDF, FILETYPE_EDF
from ._extensions._pyedflib import open_file_writeonly, set_physical_maximum, set_patient_additional, set_digital_maximum
from ._extensions._pyedflib import set_birthdate, set_digital_minimum, set_technician, set_recording_additional, set_patientname
from ._extensions._pyedflib import set_patientcode, set_equipment, set_admincode, set_sex, set_datarecord_duration, set_number_of_annotation_signals
from ._extensions._pyedflib import set_startdatetime, set_starttime_subsecond, set_samples_per_record, set_physical_minimum, set_label, set_physical_dimension
from ._extensions._pyedflib import set_transducer, set_prefilter, write_physical_samples, close_file, write_annotation_latin1, write_annotation_utf8
from ._extensions._pyedflib import blockwrite_physical_samples, write_errors, blockwrite_digital_samples, write_digital_short_samples, write_digital_samples, blockwrite_digital_short_samples


__all__ = ['EdfWriter']


def check_is_ascii(string):
    """according to the EDF+ specifications, only ASCII chars in ordeal
    range 32...126 are allowed, where 32 is space

    https://www.edfplus.info/specs/edfplus.html#header
    """
    if not all([ord(x)>32 and ord(x)<127 for x in string]):
        warnings.warn('Invalid char: header entries should contain only ASCII'\
                      ' characters and no spaces: "{}"'.format(string))


def check_signal_header_correct(channels, i, file_type):
    """
    helper function  to check if all entries in the channel dictionary are fine.

    Will give a warning if label, transducer, dimension, prefilter are too long.

    Will throw an exception if dmin, dmax, pmin, pmax are out of bounds or would
    be truncated in such a way as that signal values would be completely off.
    """
    ch = channels[i]
    label = ch['label']

    if len(ch['label'])>16:
        warnings.warn('Label of channel {} is longer than 16 ASCII chars.'\
                      'The label will be truncated to "{}"'.format(i, ch['label'][:16] ))
    if len(ch['prefilter'])>80:
        warnings.warn('prefilter of channel {} is longer than 80 ASCII chars.'\
                      'The label will be truncated to "{}"'.format(i, ch['prefilter'][:80] ))
    if len(ch['transducer'])>80:
        warnings.warn('transducer of channel {} is longer than 80 ASCII chars.'\
                      'The label will be truncated to "{}"'.format(i, ch['transducer'][:80] ))
    if len(ch['dimension'])>80:
        warnings.warn('dimension of channel {} is longer than 8 ASCII chars.'\
                      'The label will be truncated to "{}"'.format(i, ch['dimension'][:8] ))

    # these ones actually raise an exception
    dmin, dmax = (-8388608, 8388607) if file_type in (FILETYPE_BDFPLUS, FILETYPE_BDF) else (-32768, 32767)
    if ch['digital_min']<dmin:
        raise ValueError('Digital minimum for channel {} ({}) is {},'\
                         'but minimum allowed value is {}'.format(i, label,
                                                                  ch['digital_min'],
                                                                  dmin))
    if ch['digital_max']>dmax:
        raise ValueError('Digital maximum for channel {} ({}) is {},'\
                         'but maximum allowed value is {}'.format(i, label,
                                                                  ch['digital_max'],
                                                                  dmax))


    # if we truncate the physical min before the dot, we potentitally
    # have all the signals incorrect by an order of magnitude.
    if len(str(ch['physical_min']))>8 and ch['physical_min'] < -99999999:
        raise ValueError('Physical minimum for channel {} ({}) is {}, which has {} chars, '\
                         'however, EDF+ can only save 8 chars, critical precision loss is expected, '\
                         'please convert the signals to another dimesion (eg uV to mV)'.format(i, label,
                                                                      ch['physical_min'],
                                                                      len(str(ch['physical_min']))))
    if len(str(ch['physical_max']))>8 and ch['physical_max'] > 99999999:
        raise ValueError('Physical minimum for channel {} ({}) is {}, which has {} chars, '\
                         'however, EDF+ can only save 8 chars, critical precision loss is expected, '\
                         'please convert the signals to another dimesion (eg uV to mV).'.format(i, label,
                                                                      ch['physical_max'],
                                                                      len(str(ch['physical_max']))))
    # if we truncate the physical min behind the dot, we just lose precision,
    # in this case only a warning is enough
    if len(str(ch['physical_min']))>8:
        warnings.warn('Physical minimum for channel {} ({}) is {}, which has {} chars, '\
                         'however, EDF+ can only save 8 chars, will be truncated to {}, '\
                         'some loss of precision is to be expected'.format(i, label,
                                                                      ch['physical_min'],
                                                                      len(str(ch['physical_min'])),
                                                                      str(ch['physical_min'])[:8]))
    if len(str(ch['physical_max']))>8:
        warnings.warn('Physical minimum for channel {} ({}) is {}, which has {} chars, '\
                         'however, EDF+ can only save 8 chars, will be truncated to {}, '\
                         'some loss of precision is to be expected.'.format(i, label,
                                                                      ch['physical_max'],
                                                                      len(str(ch['physical_max'])),
                                                                      str(ch['physical_max'])[:8]))



def u(x):
    return x.decode("utf_8", "strict")


def du(x):
    if isinstance(x, bytes):
        return x
    else:
        return x.encode("utf_8")


def isstr(s):
    warnings.warn("Function 'isstr' is deprecated.", DeprecationWarning, stacklevel=2)
    return isinstance(s, str)


def isbytestr(s):
    warnings.warn("Function 'isbytestr' is deprecated.", DeprecationWarning, stacklevel=2)
    return isinstance(s, bytes)


def sex2int(sex):
    if isinstance(sex, int) or sex is None:
        return sex
    elif sex.lower() in ('', 'x', 'xx', 'xxx', 'unknown', '?', '??'):
        return None
    elif sex.lower() in ("female", "woman", "f", "w"):
        return 0
    elif sex.lower() in  ("male", "man", "m"):
        return 1
    else:
        raise ValueError(f"Unknown sex: '{sex}'")


def gender2int(gender):
    warnings.warn("Function 'gender2int' is deprecated, use 'sex2int' instead.", DeprecationWarning, stacklevel=2)
    return sex2int(gender)


class ChannelDoesNotExist(Exception):
    def __init__(self, value):
        self.parameter = value

    def __str__(self):
        return repr(self.parameter)


class WrongInputSize(Exception):
    def __init__(self, value):
        self.parameter = value

    def __str__(self):
        return repr(self.parameter)


class EdfWriter:
    def __exit__(self, exc_type, exc_val, ex_tb):
        self.close()

    def __enter__(self):
        return self

    def __del__(self):
        self.close()

    def __init__(self, file_name, n_channels,
                 file_type=FILETYPE_EDFPLUS):
        """Initialises an EDF file at file_name.
        file_type is one of
            edflib.FILETYPE_EDFPLUS
            edflib.FILETYPE_BDFPLUS
        n_channels is the number of channels without the annotation channel

        channel_info should be a
        list of dicts, one for each channel in the data. Each dict needs
        these values:

            'label' : channel label (string, <= 16 characters, must be unique)
            'dimension' : physical dimension (e.g., mV) (string, <= 8 characters)
            'sample_rate' : sample frequency in hertz (int). Deprecated: use 'sample_frequency' instead.
            'sample_frequency' : number of samples per record (int)
            'physical_max' : maximum physical value (float)
            'physical_min' : minimum physical value (float)
            'digital_max' : maximum digital value (int, -2**15 <= x < 2**15)
            'digital_min' : minimum digital value (int, -2**15 <= x < 2**15)
        """
        self.path = file_name
        self.file_type = file_type
        self.patient_name = ''
        self.patient_code = ''
        self.technician = ''
        self.equipment = ''
        self.recording_additional = ''
        self.patient_additional = ''
        self.admincode = ''
        self.sex = None
        self.recording_start_time = datetime.now().replace(microsecond=0)

        self.birthdate = ''
        self.record_duration = 1  # length of one data record in seconds
        self.number_of_annotations = 1 if file_type in [FILETYPE_EDFPLUS, FILETYPE_BDFPLUS] else 0
        self.n_channels = n_channels
        self.channels = []
        self.sample_buffer = []
        for i in np.arange(self.n_channels):
            if self.file_type == FILETYPE_BDFPLUS or self.file_type == FILETYPE_BDF:
                self.channels.append({'label': f'ch{i}', 'dimension': 'mV', 'sample_rate': 100,
                                      'sample_frequency': None, 'physical_max': 1.0, 'physical_min': -1.0,
                                      'digital_max': 8388607,'digital_min': -8388608,
                                      'prefilter': '', 'transducer': ''})
            elif self.file_type == FILETYPE_EDFPLUS or self.file_type == FILETYPE_EDF:
                self.channels.append({'label': f'ch{i}', 'dimension': 'mV', 'sample_rate': 100,
                                      'sample_frequency': None, 'physical_max': 1.0, 'physical_min': -1.0,
                                      'digital_max': 32767, 'digital_min': -32768,
                                      'prefilter': '', 'transducer': ''})

                self.sample_buffer.append([])
        self.handle = open_file_writeonly(self.path, self.file_type, self.n_channels)
        if (self.handle < 0):
            raise OSError(write_errors[self.handle])
        self._enforce_record_duration = False

    def update_header(self):
        """
        Updates header to edffile struct
        """
        # some checks that warn users if header fields exceed 80 chars
        patient_ident = len(self.patient_code) + len(self.patient_name) \
                        + len(self.patient_additional) + 3 + 1 + 11 # 3 spaces 1 sex 11 birthdate
        record_ident = len(self.equipment) + len(self.technician) \
                       + len(self.admincode) + len(self.recording_additional) \
                       + len('Startdate') + 3 + 11 # 3 spaces 11 birthdate

        if patient_ident>80:
            warnings.warn('Patient code, name, sex and birthdate combined must not be larger than 80 chars. ' +
                          f'Currently has len of {patient_ident}. See https://www.edfplus.info/specs/edfplus.html#additionalspecs')
        if record_ident>80:
            warnings.warn('Equipment, technician, admincode and recording_additional combined must not be larger than 80 chars. ' +
                          f'Currently has len of {record_ident}. See https://www.edfplus.info/specs/edfplus.html#additionalspecs')

        # all data records (i.e. blocks of data of a channel) have one singular
        # length in seconds. If there are different sampling frequencies for
        # the channels, we need to find a common denominator so that all sample
        # frequencies can be represented accurately.
        # this can be overwritten by explicitly calling setDatarecordDuration
        self._calculate_optimal_record_duration()

        set_technician(self.handle, du(self.technician))
        set_recording_additional(self.handle, du(self.recording_additional))
        set_patientname(self.handle, du(self.patient_name))
        set_patientcode(self.handle, du(self.patient_code))
        set_patient_additional(self.handle, du(self.patient_additional))
        set_equipment(self.handle, du(self.equipment))
        set_admincode(self.handle, du(self.admincode))
        set_sex(self.handle, sex2int(self.sex))

        set_datarecord_duration(self.handle, self.record_duration)
        set_number_of_annotation_signals(self.handle, self.number_of_annotations)
        set_startdatetime(self.handle, self.recording_start_time.year, self.recording_start_time.month,
                          self.recording_start_time.day, self.recording_start_time.hour,
                          self.recording_start_time.minute, self.recording_start_time.second)
        # subseconds are noted in nanoseconds, so we multiply by 100
        if self.recording_start_time.microsecond>0:
            set_starttime_subsecond(self.handle, self.recording_start_time.microsecond*100)
        if isinstance(self.birthdate, str):
            if self.birthdate != '':
                birthday = datetime.strptime(self.birthdate, '%d %b %Y').date()
                set_birthdate(self.handle, birthday.year, birthday.month, birthday.day)
        else:
            set_birthdate(self.handle, self.birthdate.year, self.birthdate.month, self.birthdate.day)

        for i in np.arange(self.n_channels):
            check_signal_header_correct(self.channels, i, self.file_type)
            set_samples_per_record(self.handle, i, self.get_smp_per_record(i))
            set_physical_maximum(self.handle, i, self.channels[i]['physical_max'])
            set_physical_minimum(self.handle, i, self.channels[i]['physical_min'])
            set_digital_maximum(self.handle, i, self.channels[i]['digital_max'])
            set_digital_minimum(self.handle, i, self.channels[i]['digital_min'])
            set_label(self.handle, i, du(self.channels[i]['label']))
            set_physical_dimension(self.handle, i, du(self.channels[i]['dimension']))
            set_transducer(self.handle, i, du(self.channels[i]['transducer']))
            set_prefilter(self.handle, i, du(self.channels[i]['prefilter']))

    def setHeader(self, fileHeader):
        """
        Sets the file header
        """
        self.technician = fileHeader["technician"]
        self.recording_additional = fileHeader["recording_additional"]
        self.patient_name = fileHeader["patientname"]
        self.patient_additional = fileHeader["patient_additional"]
        self.patient_code = fileHeader["patientcode"]
        self.equipment = fileHeader["equipment"]
        self.admincode = fileHeader["admincode"]
        self.sex = fileHeader["sex"]
        self.recording_start_time = fileHeader["startdate"]
        self.birthdate = fileHeader["birthdate"]
        self.update_header()

    def setSignalHeader(self, edfsignal, channel_info):
        """
        Sets the parameter for signal edfsignal.

        channel_info should be a dict with
        these values:

            'label' : channel label (string, <= 16 characters, must be unique)
            'dimension' : physical dimension (e.g., mV) (string, <= 8 characters)
            'sample_rate' : sample frequency in hertz (int). Deprecated: use 'sample_frequency' instead.
            'sample_frequency' : number of samples per record (int)
            'physical_max' : maximum physical value (float)
            'physical_min' : minimum physical value (float)
            'digital_max' : maximum digital value (int, -2**15 <= x < 2**15)
            'digital_min' : minimum digital value (int, -2**15 <= x < 2**15)
        """
        if edfsignal < 0 or edfsignal > self.n_channels:
            raise ChannelDoesNotExist(edfsignal)
        self.channels[edfsignal].update(channel_info)
        self.update_header()

    def setSignalHeaders(self, signalHeaders):
        """
        Sets the parameter for all signals

        Parameters
        ----------
        signalHeaders : array_like
            containing dict with
                'label' : str
                          channel label (string, <= 16 characters, must be unique)
                'dimension' : str
                          physical dimension (e.g., mV) (string, <= 8 characters)
                'sample_rate' :
                          sample frequency in hertz (int). Deprecated: use 'sample_frequency' instead.
                'sample_frequency' : int
                          number of samples per record
                'physical_max' : float
                          maximum physical value
                'physical_min' : float
                         minimum physical value
                'digital_max' : int
                         maximum digital value (-2**15 <= x < 2**15)
                'digital_min' : int
                         minimum digital value (-2**15 <= x < 2**15)
        """
        for edfsignal in np.arange(self.n_channels):
            self.channels[edfsignal].update(signalHeaders[edfsignal])
        self.update_header()

    def setTechnician(self, technician):
        """
        Sets the technicians name to `technician`.

        Notes
        -----
        This function is optional and can be called only after opening a file in writemode and before the first sample write action.
        """
        check_is_ascii(technician)
        self.technician = technician
        self.update_header()

    def setRecordingAdditional(self, recording_additional):
        """
        Sets the additional recordinginfo

        Notes
        -----
        This function is optional and can be called only after opening a file in writemode and before the first sample write action.
        """
        check_is_ascii(recording_additional)
        self.recording_additional = recording_additional
        self.update_header()

    def setPatientName(self, patient_name):
        """
        Sets the patientname to `patient_name`.

        Notes
        -----
        This function is optional and can be called only after opening a file in writemode and before the first sample write action.
        """
        check_is_ascii(patient_name)
        self.patient_name = patient_name
        self.update_header()

    def setPatientCode(self, patient_code):
        """
        Sets the patientcode to `patient_code`.

        Notes
        -----
        This function is optional and can be called only after opening a file in writemode and before the first sample write action.
        """
        check_is_ascii(patient_code)
        self.patient_code = patient_code
        self.update_header()

    def setPatientAdditional(self, patient_additional):
        """
        Sets the additional patientinfo to `patient_additional`.

        Notes
        -----
        This function is optional and can be called only after opening a file in writemode and before the first sample write action.
        """
        check_is_ascii(patient_additional)
        self.patient_additional = patient_additional
        self.update_header()

    def setEquipment(self, equipment):
        """
        Sets the name of the param equipment used during the acquisition.
        This function is optional and can be called only after opening a file in writemode and before the first sample write action.

        Parameters
        ----------
        equipment : str
            Describes the measurement equpipment

        """
        check_is_ascii(equipment)
        self.equipment = equipment
        self.update_header()

    def setAdmincode(self, admincode):
        """
        Sets the admincode.

        This function is optional and can be called only after opening a file in writemode and before the first sample write action.

        Parameters
        ----------
        admincode : str
            admincode which is written into the header

        """
        check_is_ascii(admincode)
        self.admincode = admincode
        self.update_header()

    def setSex(self, sex):
        """
        Sets the sex. Due to the edf specifications, only binary assignment is possible.
        This function is optional and can be called only after opening a file in writemode and before the first sample write action.

        Parameters
        ----------
        sex : int
            1 is male, 0 is female
        """
        self.sex = sex2int(sex)
        self.update_header()

    def setGender(self, gender):
        warnings.warn("Function 'setGender' is deprecated, use 'setSex' instead.", DeprecationWarning, stacklevel=2)
        self.sex = sex2int(gender)
        self.update_header()

    def setDatarecordDuration(self, record_duration):
        """
        Sets the datarecord duration. The default value is 1 second.
        The datarecord duration must be in the range 0.001 to 60  seconds.
        Usually, the datarecord duration is calculated automatically to
        ensure that all sample frequencies are representable, nevertheless,
        you can overwrite the datarecord duration manually. This can, however,
        lead to unexpected side-effects in the sample frequency calculations.

        Returns 0 on success, otherwise -1.

        Parameters
        ----------
        record_duration : float
            Sets the datarecord duration in units of seconds

        Notes
        -----
        This function is NOT REQUIRED but can be called after opening a file
        in writemode and before the first sample write action. This function
        can be used when you want to use a samplefrequency which is not an
        integer. For example, if you want to use a samplerate of 0.5 Hz, set
        the samplefrequency to 5 Hz and the datarecord duration to 10 seconds.
        Do not use this function, except when absolutely necessary!
        """
        self._enforce_record_duration = True
        self.record_duration = record_duration
        self.update_header()

    def set_number_of_annotation_signals(self, number_of_annotations):
        """
        Sets the number of annotation signals. The default value is 1
        This function is optional and can be called only after opening a file in writemode
        and before the first sample write action
        Normally you don't need to change the default value. Only when the number of annotations
        you want to write is more than the number of seconds of the duration of the recording, you can use
        this function to increase the storage space for annotations
        Minimum is 1, maximum is 64

        Parameters
        ----------
        number_of_annotations : integer
            Sets the number of annotation signals
        """
        number_of_annotations = max((min((int(number_of_annotations), 64)), 1))
        self.number_of_annotations = number_of_annotations
        self.update_header()

    def setStartdatetime(self, recording_start_time):
        """
        Sets the recording start Time

        Parameters
        ----------
        recording_start_time: datetime object
            Sets the recording start Time
        """
        if not isinstance(recording_start_time, datetime):
            recording_start_time = datetime.strptime(recording_start_time,"%d %b %Y %H:%M:%S")
        self.recording_start_time = recording_start_time
        self.update_header()

    def setBirthdate(self, birthdate):
        """
        Sets the birthdate.

        Parameters
        ----------
        birthdate: date object from datetime

        Examples
        --------
        >>> import pyedflib
        >>> from datetime import datetime, date
        >>> f = pyedflib.EdfWriter('test.bdf', 1, file_type=pyedflib.FILETYPE_BDFPLUS)
        >>> f.setBirthdate(date(1951, 8, 2))
        >>> f.close()

        Notes
        -----
        This function is optional and can be called only after opening a file in writemode and before the first sample write action.
        """
        if isinstance(birthdate, str):
            birthdate = datetime.strptime(birthdate, "%d.%m.%Y")
        self.birthdate = birthdate
        self.update_header()

    def setSamplefrequency(self, edfsignal, samplefrequency):
        """
        Sets the samplefrequency of signal edfsignal.

        Parameters
        ----------
        edfsignal: index number of signal
        samplefrequency: int or float stating the sampling frequency in Hz.
                        internally, EDF stores the sampling frequency by
                        setting the smp_per_record and the record_duration.
                        That means, the sampling frequency is not stored
                        explicitly.

        Notes
        -----
        This function is required for every signal and can be called only after
        opening a file in writemode and before the first sample write action.
        """
        if edfsignal < 0 or edfsignal > self.n_channels:
            raise ChannelDoesNotExist(edfsignal)

        self.channels[edfsignal]['sample_frequency'] = samplefrequency
        self.update_header()

    def setPhysicalMaximum(self, edfsignal, physical_maximum):
        """
        Sets the physical_maximum of signal edfsignal.

        Parameters
        ----------
        edfsignal: int
            signal number
        physical_maximum: float
            Sets the physical maximum

        Notes
        -----
        This function is required for every signal and can be called only after opening a file in writemode and before the first sample write action.
        """
        if edfsignal < 0 or edfsignal > self.n_channels:
            raise ChannelDoesNotExist(edfsignal)
        self.channels[edfsignal]['physical_max'] = physical_maximum
        self.update_header()

    def setPhysicalMinimum(self, edfsignal, physical_minimum):
        """
        Sets the physical_minimum of signal edfsignal.

        Parameters
        ----------
        edfsignal: int
            signal number
        physical_minimum: float
            Sets the physical minimum

        Notes
        -----
        This function is required for every signal and can be called only after opening a file in writemode and before the first sample write action.
        """
        if (edfsignal < 0 or edfsignal > self.n_channels):
            raise ChannelDoesNotExist(edfsignal)
        self.channels[edfsignal]['physical_min'] = physical_minimum
        self.update_header()

    def setDigitalMaximum(self, edfsignal, digital_maximum):
        """
        Sets the maximum digital value of signal edfsignal.
        Usually, the value 32767 is used for EDF+ and 8388607 for BDF+.

        Parameters
        ----------
        edfsignal : int
            signal number
        digital_maximum : int
            Sets the maximum digital value

        Notes
        -----
        This function is optional and can be called only after opening a file in writemode and before the first sample write action.
        """
        if (edfsignal < 0 or edfsignal > self.n_channels):
            raise ChannelDoesNotExist(edfsignal)
        self.channels[edfsignal]['digital_max'] = digital_maximum
        self.update_header()

    def setDigitalMinimum(self, edfsignal, digital_minimum):
        """
        Sets the minimum digital value of signal edfsignal.
        Usually, the value -32768 is used for EDF+ and -8388608 for BDF+. Usually this will be (-(digital_maximum + 1)).

        Parameters
        ----------
        edfsignal : int
            signal number
        digital_minimum : int
            Sets the minimum digital value

        Notes
        -----
        This function is optional and can be called only after opening a file in writemode and before the first sample write action.
        """
        if (edfsignal < 0 or edfsignal > self.n_channels):
            raise ChannelDoesNotExist(edfsignal)
        self.channels[edfsignal]['digital_min'] = digital_minimum
        self.update_header()

    def setLabel(self, edfsignal, label):
        """
        Sets the label (name) of signal edfsignal ("FP1", "SaO2", etc.).

        Parameters
        ----------
        edfsignal : int
            signal number on which the label should be changed
        label : str
            signal label

        Notes
        -----
        This function is recommended for every signal and can be called only after opening a file in writemode and before the first sample write action.
        """
        if (edfsignal < 0 or edfsignal > self.n_channels):
            raise ChannelDoesNotExist(edfsignal)
        self.channels[edfsignal]['label'] = label
        self.update_header()

    def setPhysicalDimension(self, edfsignal, physical_dimension):
        """
        Sets the physical dimension of signal edfsignal ("uV", "BPM", "mA", "Degr.", etc.)

        :param edfsignal: int
        :param physical_dimension: str

        Notes
        -----
        This function is recommended for every signal and can be called only after opening a file in writemode and before the first sample write action.
        """
        if edfsignal < 0 or edfsignal > self.n_channels:
            raise ChannelDoesNotExist(edfsignal)
        self.channels[edfsignal]['dimension'] = physical_dimension
        self.update_header()

    def setTransducer(self, edfsignal, transducer):
        """
        Sets the transducer of signal edfsignal

        :param edfsignal: int
        :param transducer: str

        Notes
        -----
        This function is optional for every signal and can be called only after opening a file in writemode and before the first sample write action.
        """
        if (edfsignal < 0 or edfsignal > self.n_channels):
            raise ChannelDoesNotExist(edfsignal)
        self.channels[edfsignal]['transducer'] = transducer
        self.update_header()

    def setPrefilter(self, edfsignal, prefilter):
        """
        Sets the prefilter of signal edfsignal ("HP:0.1Hz", "LP:75Hz N:50Hz", etc.)

        :param edfsignal: int
        :param prefilter: str

        Notes
        -----
        This function is optional for every signal and can be called only after opening a file in writemode and before the first sample write action.
        """
        if edfsignal < 0 or edfsignal > self.n_channels:
            raise ChannelDoesNotExist(edfsignal)
        self.channels[edfsignal]['prefilter'] = prefilter
        self.update_header()

    def writePhysicalSamples(self, data):
        """
        Writes n physical samples (uV, mA, Ohm) belonging to one signal where n
        is the samplefrequency of the signal.

        data_vec belonging to one signal. The size must be the samplefrequency of the signal.

        Notes
        -----
        Writes n physical samples (uV, mA, Ohm) from data_vec belonging to one signal where n
        is the samplefrequency of the signal. The physical samples will be converted to digital
        samples using the values of physical maximum, physical minimum, digital maximum and digital
        minimum. The number of samples written is equal to the samplefrequency of the signal.
        Call this function for every signal in the file. The order is important! When there are 4
        signals in the file, the order of calling this function must be: signal 0, signal 1, signal 2,
        signal 3, signal 0, signal 1, signal 2, etc.

        All parameters must be already written into the bdf/edf-file.
        """
        return write_physical_samples(self.handle, data)

    def writeDigitalSamples(self, data):
        return write_digital_samples(self.handle, data)

    def writeDigitalShortSamples(self, data):
        return write_digital_short_samples(self.handle, data)

    def blockWritePhysicalSamples(self, data):
        """
        Writes physical samples (uV, mA, Ohm)
        must be filled with samples from all signals
        where each signal has n samples which is the samplefrequency of the signal.

        data_vec belonging to one signal. The size must be the samplefrequency of the signal.

        Notes
        -----
        buf must be filled with samples from all signals, starting with signal 0, 1, 2, etc.
        one block equals one second
        The physical samples will be converted to digital samples using the
        values of physical maximum, physical minimum, digital maximum and digital minimum
        The number of samples written is equal to the sum of the samplefrequencies of all signals
        Size of buf should be equal to or bigger than sizeof(double) multiplied by the sum of the samplefrequencies of all signals
        Returns 0 on success, otherwise -1

        All parameters must be already written into the bdf/edf-file.
        """
        return blockwrite_physical_samples(self.handle, data)

    def blockWriteDigitalSamples(self, data):
        return blockwrite_digital_samples(self.handle, data)

    def blockWriteDigitalShortSamples(self, data):
        return blockwrite_digital_short_samples(self.handle, data)

    def writeSamples(self, data_list, digital = False):
        """
        Writes physical samples (uV, mA, Ohm) from data belonging to all signals
        The physical samples will be converted to digital samples using the values
        of physical maximum, physical minimum, digital maximum and digital minimum.
        if the samplefrequency of all signals are equal, then the data could be
        saved into a matrix with the size (N,signals) If the samplefrequency
        is different, then sample_freq is a vector containing all the different
        samplefrequencys. The data is saved as list. Each list entry contains
        a vector with the data of one signal.

        If digital is True, digital signals (as directly from the ADC) will be expected.
        (e.g. int16 from 0 to 2048)

        All parameters must be already written into the bdf/edf-file.
        """
        there_are_blank_sample_frequencies = any([channel.get('sample_frequency') is None
                                                 for channel in self.channels])
        if there_are_blank_sample_frequencies:
            warnings.warn("The 'sample_rate' parameter is deprecated. Please use "
                          "'sample_frequency' instead.", DeprecationWarning)

        if (len(data_list)) == 0:
            raise WrongInputSize('Data list is empty')
        if (len(data_list) != len(self.channels)):
            raise WrongInputSize('Number of channels ({}) \
             unequal to length of data ({})'.format(len(self.channels), len(data_list)))

        # Check for F-contiguous arrays
        if not all(s.flags.c_contiguous for s in data_list):
            warnings.warn('signals are in Fortran order. Will automatically '
                          'transfer to C order for compatibility with edflib.')
            data_list = np.ascontiguousarray(data_list)

        if digital:
            if any([not np.issubdtype(a.dtype, np.integer) for a in data_list]):
                raise TypeError('Digital = True requires all signals in int')

        # Check that all channels have different physical_minimum and physical_maximum
        for chan in self.channels:
            assert chan['physical_min'] != chan['physical_max'], \
            'In chan {} physical_min {} should be different from '\
            'physical_max {}'.format(chan['label'], chan['physical_min'], chan['physical_max'])

        ind = []
        notAtEnd = True
        for i in np.arange(len(data_list)):
            ind.append(0)

        sampleLength = 0
        smp_per_record = np.zeros(len(data_list), dtype=np.int32)
        for i in np.arange(len(data_list)):
            smp_per_record[i] = self.get_smp_per_record(i)

            if (np.size(data_list[i]) < ind[i] + smp_per_record[i]):
                notAtEnd = False
            sampleLength += smp_per_record[i]

        dataRecord = np.array([], dtype=np.int32 if digital else None)

        while notAtEnd:
            del dataRecord
            dataRecord = np.array([], dtype=np.int32 if digital else None)
            for i in np.arange(len(data_list)):
                dataRecord = np.append(dataRecord, data_list[i][int(ind[i]):int(ind[i]+smp_per_record[i])])
                ind[i] += smp_per_record[i]
            if digital:
                success = self.blockWriteDigitalSamples(dataRecord)
            else:
                success = self.blockWritePhysicalSamples(dataRecord)

            if success < 0:
                raise OSError(f'Unknown error while calling blockWriteSamples: {success}')

            for i in np.arange(len(data_list)):
                if (np.size(data_list[i]) < ind[i] + smp_per_record[i]):
                    notAtEnd = False


        for i in np.arange(len(data_list)):
            lastSamples = np.zeros(smp_per_record[i], dtype=np.int32 if digital else None)
            lastSampleInd = int(np.max(data_list[i].shape) - ind[i])
            lastSampleInd = int(np.min((lastSampleInd, smp_per_record[i])))
            if lastSampleInd > 0:
                lastSamples[:lastSampleInd] = data_list[i][-lastSampleInd:]
                if digital:
                    success = self.writeDigitalSamples(lastSamples)
                else:
                    success = self.writePhysicalSamples(lastSamples)

                if success<0:
                    raise OSError(f'Unknown error while calling writeSamples: {success}')

    def writeAnnotation(self, onset_in_seconds, duration_in_seconds, description, str_format='utf_8'):
        """
        Writes an annotation/event to the file
        """
        if self.file_type in [FILETYPE_EDF, FILETYPE_BDF]:
            raise TypeError('Trying to write annotation to EDF/BDF, must use EDF+/BDF+')

        if isinstance(duration_in_seconds, bytes):
            duration_in_seconds = float(duration_in_seconds)

        if str_format == 'utf_8':
            if duration_in_seconds >= 0:
                return write_annotation_utf8(self.handle, np.round(onset_in_seconds*10000).astype(np.int64), np.round(duration_in_seconds*10000).astype(int), du(description))
            else:
                return write_annotation_utf8(self.handle, np.round(onset_in_seconds*10000).astype(np.int64), -1, du(description))
        else:
            if duration_in_seconds >= 0:
                return write_annotation_latin1(self.handle, np.round(onset_in_seconds*10000).astype(np.int64), np.round(duration_in_seconds*10000).astype(int), u(description).encode('latin1'))
            else:
                return write_annotation_latin1(self.handle, np.round(onset_in_seconds*10000).astype(np.int64), -1, u(description).encode('latin1'))

    def close(self):
        """
        Closes the file.
        """
        close_file(self.handle)
        self.handle = -1

    def get_smp_per_record(self, ch_idx):
        """
        gets the calculated number of samples that need to be fit into one
        record (i.e. window/block of data) with the given record duration.
        """
        fs = self._get_sample_frequency(ch_idx)
        if fs is None: return None

        record_duration = self.record_duration
        smp_per_record = fs*record_duration

        if not np.isclose(np.round(smp_per_record), np.round(smp_per_record, 6)):
            warnings.warn(f'Sample frequency {fs} can not be represented accurately. \n' +
                          f'smp_per_record={smp_per_record}, record_duration={record_duration} seconds,' +
                          f'calculated sample_frequency will be {np.round(smp_per_record)/record_duration}')
        return int(np.round(smp_per_record))


    def _calculate_optimal_record_duration(self):
        """
        calculate optimal denominator (record duration in seconds)
        for all sample frequencies such that smp_per_record is an integer
        for all channels.

        If all sampling frequencies are integers, this will simply be 1.
        """
        if self._enforce_record_duration: return
        allint = lambda int_list: all([n==int(n) for n in int_list])
        all_fs = [self._get_sample_frequency(i) for i,_ in enumerate(self.channels)]

        # calculate the optimal record duration to represent all frequencies.
        # this is achieved when fs*duration=int, i.e. the number of samples
        # in one data record can be represented by an int (smp_per_record)
        # if all sampling frequencies are ints, this will be simply 1
        # for now this brute force solution should cover 99% of cases.
        # TODO: optimize this process

        record_duration = 0
        for i in range(1, 60):
            if allint([x*i for x in all_fs]):
                record_duration = i
                break
        assert record_duration>0, f'cannot accurately represent sampling frequencies with data record durations between 1-60s: {all_fs}'
        assert record_duration<=60, 'record duration must be below 60 seconds'
        self.record_duration = record_duration

    def _get_sample_frequency(self, channelIndex):
        # Temporary conditional assignment while we deprecate 'sample_rate' as a channel attribute
        # in favor of 'sample_frequency', supporting the use of either to give
        # users time to switch to the new interface.
        if 'sample_rate' in self.channels[channelIndex]:
            warnings.warn("`sample_rate` is deprecated and will be removed in a future release. \
                          Please use `sample_frequency` instead", DeprecationWarning)
        return (self.channels[channelIndex]['sample_rate']
                if self.channels[channelIndex].get('sample_frequency') is None
                else self.channels[channelIndex]['sample_frequency'])
